def mass_models(mass_distribution,n_galaxies, p1, lens_centered, path):
    if mass_distribution == "SIS":
            # Initialize l as an empty string
            model = f"alpha {p1} 0 0 0.0 0.0 0 0.0 0 0 1\n"
            if n_galaxies > 1:
                for n in range(1, n_galaxies):
                    ra, dec = lens_centered[n]
                    # Create the line with p1/2 and the current ra, dec
                    model += f"alpha {p1 / 2} {ra} {dec} 0.0 0.0 0 0.0 0 0 1\n"
                    #l += line  # Concatenate the line to l
            # Add the fixed line
            model += "1 1 1 0 0 0 0 0 0 0\n"
            if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
            # Add the optimize line with the given path
            model += f"optimize {path}/final_step\n"
    elif  mass_distribution in ["SIE","POW"]:
            # Initialize l as an empty string
            model = f"alpha {p1} 0 0 0.003 10.0 0 0.0 0 0 1\n"
            if n_galaxies > 1:
                for n in range(1, n_galaxies):
                    ra, dec = lens_centered[n]
                    # Create the line with p1/2 and the current ra, dec
                    model += f"alpha {p1 / 2} {ra} {dec} 0.0 0.0 0 0.0 0 0 1\n"
                    #l += line  # Concatenate the line to l
            # Add the fixed line
            model += "1 0 0 1 1 0 0 0 0 0\n"
            if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
            model += f"varyone 1 5 -90.0 90.0 19 {path}/step_1\n"
            model += f"\nsetlens {path}/step_1.start\nchangevary 1"#heer maybe is the number of galaxies  
            model += f"\n1 0 0 1 1 0 0 0 0 0 \n"
            if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 1 1 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
            model += f"optimize {path}/step_2\n"
            if mass_distribution=="SIE":
                model += f"\nset chimode= 1\nsetlens {path}/step_2.start\nchangevary 1\n1 1 1 1 1 0 0 0 0 0\n"
            elif mass_distribution=="POW":
                model += f"\nset chimode= 1\nsetlens {path}/step_2.start\nchangevary 1\n1 1 1 1 1 0 0 0 0 1\n"
            if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
            model += f"\noptimize {path}/final_step\n"
    elif "SIS+shear" == mass_distribution:
        model = f"alpha {p1} 0 0 0.0 0.0 0.003 10.0 0 0 1\n"
        if n_galaxies > 1:
                for n in range(1, n_galaxies):
                    ra, dec = lens_centered[n]
                    # Create the line with p1/2 and the current ra, dec
                    model += f"alpha {p1 / 2} {ra} {dec} 0.0 0.0 0 0.0 0 0 1\n"
                    #l += line  # Concatenate the line to l
        model+= "1 0 0 0 0 1 1 0 0 0\n"
        if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
        model+= f"varyone 1 7 -90.0 90.0 19 {path}/step_1\n"
        model+= f"setlens {path}/step_1.start\n"
        model+= f"changevary 1 \n1 0 0 0 0 1 1 0 0 0\n"
        if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
        model+= f"\noptimize {path}/step_2\n"
        model += f"set chimode= 1\nsetlens {path}/step_2.start\n"
        model += f"changevary 1\n"
        model += "1 1 1 0 0 1 1 0 0 0\n"
        if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
        model += f"optimize {path}/final_step\n"
    elif mass_distribution in ["SIE+shear","POW+shear"]:
        model = f"alpha {p1} 0 0 0.03 10.0 0 0.0 0 0 1\n"
        if n_galaxies > 1:
                for n in range(1, n_galaxies):
                    ra, dec = lens_centered[n]
                    # Create the line with p1/2 and the current ra, dec
                    model += f"alpha {p1} {ra} {dec} 0.0 0.0 0 0.0 0 0 1\n"
                    #l += line  # Concatenate the line to l
        model+= "1 0 0 1 1 0 0 0 0 0\n"
        model+= f"varytwo 1 5 -90.0 90.0 19 1 7 -90.0 90.0 19 {path}/step_1\n"
        model+= f"setlens {path}/step_1.start\n"
        model+= f"changevary 1 \n1 1 1 1 1 0 0 0 0 0\n"
        if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
        model+= f"\noptimize {path}/step_2\n"
        model += f"set chimode= 1\nsetlens {path}/step_2.start\n"
        model += f"changevary 1\n"
        if "SIE+shear" ==mass_distribution:
            model += "1 1 1 1 1 1 1 0 0 0\n"
        elif "POW+shear" ==mass_distribution:
            model += "1 1 1 1 1 1 1 0 0 1\n"
        if n_galaxies > 1:
                for _ in range(1, n_galaxies):
                    model += "1 0 0 0 0 0 0 0 0 0\n"  # Concatenate the fixed line multiple times
        model += f"optimize {path}/final_step\n"
    else: 
        raise print(mass_distribution,"not avalaible")
    model += f"set plotmode    = 2\nkapgam 3 {path}/kappa_gamma.dat\nplotcrit {path}/critical_caustic.dat\nplotgrid {path}/grid.dat\ncalcRein 10 {path}/RE.dat\nquit"    
    return model


import numpy as np 
import os 
import pandas as pd 
import warnings
import json
import matplotlib.pyplot as plt 
from copy import deepcopy
from .functions import linear_model,linear_func
from scipy.integrate import quad
from matplotlib.widgets import Button,RangeSlider
import pickle

module_dir = os.path.dirname(os.path.abspath(__file__))
windows_rest_frame = os.path.join(module_dir,"rest_frame_windows","rest_frame_windows.json")
with open(windows_rest_frame, "r") as f:
    windows_rest_frame = json.load(f)
def convert_none_to_nan(item):
    if item is None:
        return np.nan
    elif isinstance(item, list):
        return [convert_none_to_nan(x) for x in item]
    elif isinstance(item, dict):
        return {key: convert_none_to_nan(value) if isinstance(value[0],str) else convert_none_to_nan(value) for key,value in item.items()}
    else:
        return item




class WindowAnalysis:
    def __init__(self,results,zs=0.0,rest_frame=True,save_name = None,obj_name=None,path_previous_results=None):
        """_summary_

        Args:
            results (_type_): _description_
            zs (_type_, optional): _description_. Defaults to None.
            rest_frame (bool, optional): _description_. Defaults to True.
        """
        assert isinstance(zs,float) or isinstance(zs,int) , "zs have to be float or int"
        self.zs = zs
        self.save_name = save_name
        self._previous_results =  self._read_previous_results(path_previous_results)
        
        if not self.save_name:
            self.save_name = "flux_cont_core.csv"
        if self.zs == 0:
            print("Warning: zs set to default value 0.0.")
        self.pre_define_windows = convert_none_to_nan(windows_rest_frame)
        if not rest_frame:
            for i in [ "left_range","right_range","core_range"]:
                self.pre_define_windows[i] = (np.array(self.pre_define_windows[i])*(1+self.zs)).tolist()
        self.pre_define_windows = pd.DataFrame(self.pre_define_windows)
        if isinstance(results,str):
            print("WORK IN PROGRESS")
        elif isinstance(results,dict):
            results = deepcopy(results)
            band_list = []
            self.spectra_dict = {}
            for obj,value in results.items():
                if "G" in obj:
                    print("We will discard the spectra's galaxy ")
                    continue
                band = value.get("band",obj.split("_")[1])
                band_list.append(band)
                if band not in self.spectra_dict.keys():
                    self.spectra_dict[band] = {}
                    for i in ["wavelength","flux","error","obj"]:
                        self.spectra_dict[band][i] = []
                wavelength = value.get("wavelength")
                if rest_frame:
                    wavelength = wavelength/(1+self.zs)
                self.spectra_dict[band]["wavelength"].append(wavelength)
                flux = value.get("flux")
                self.spectra_dict[band]["flux"].append(flux)
                self.spectra_dict[band]["error"].append(value.get("error",np.ones_like(flux)))
                self.spectra_dict[band]["obj"].append(obj)
    
    def __call__(self,**kwargs):
        lines = self.pre_define_windows["line_name"].values
        for line_name in lines:
            #print(line_name)
            row = self.pre_define_windows[self.pre_define_windows["line_name"]==line_name]
            center_window = np.mean(row["core_range"].values[0])
            window = [center_window-500,center_window+500]
            lr_init = row["right_range"].values[0]
            lc_init = row["left_range"].values[0]
            core_init = row["core_range"].values[0]
            #print(center_window,lr_init,lr_init,lc_init,core_init)
            if self._previous_results is not None and line_name in self._previous_results.line_name.values:
                row = self._previous_results[self._previous_results.line_name == line_name].iloc[0].to_dict() #assuming all the rows share the same values
                row.update({["X","Y","objs"][n]: value for n,value in enumerate(np.array(self.spectra_dict[row.band]["wavelength"]), np.array(self.spectra_dict[row.band]["flux"]),self.spectra_dict[row.band]["obj"])})
                self.interactive_plot(**row)
            for band in self.spectra_dict.keys():
                X,Y,objs = np.array(self.spectra_dict[band]["wavelength"]), np.array(self.spectra_dict[band]["flux"]),self.spectra_dict[band]["obj"]
                if center_window < np.min(X) or center_window > np.max(X):
                    continue
                else:
                    #print( center_window,np.min(X),np.max(X))
                    self.interactive_plot(X,Y,objs,center_window,window,lr_init,lc_init,core_init,band,line_name)
    
    def interactive_plot(self,X,Y,objs,center_window,window,lr_init,lc_init,core_init,band,line_name):
        #X,Y,objs = np.array(self.spectra_dict[band]["wavelength"]), np.array(self.spectra_dict[band]["flux"]),self.spectra_dict[band]["obj"]
        #if center_window < np.min(X[0]) or center_window > np.max(X[0]):
         #   return "cant do nothing"
        #else:
            w_mask = (X[0]>=min(window)) & (X[0]<=max(window)) #soft coming sooon 
            Y_local = Y[:,w_mask]
            q2 = np.percentile(Y_local, 89)
            q3 = np.percentile(Y_local, 95) 
            q4 = np.percentile(Y_local, 99.97) 
            max_local = np.max(Y_local)
            min_local = np.min(Y_local)
            aspect_ratio = 1.5
            fig = plt.figure(figsize=(20, 15 / 1.5))
            grid = plt.GridSpec(2, 2, width_ratios=[2, 2], height_ratios=[3, 1], hspace=0.4)
            Lp = plt.subplot(grid[0, 0])
            Rp = plt.subplot(grid[0, 1])
            bbox_Lp = Lp.get_position()
            bbox_Rp = Rp.get_position()
            gap_left = bbox_Lp.x0/4 + bbox_Lp.width
            #the val init aqui luego 
            #line -> todo el resto adentro ? lo q seria un doble looop?
            #if self._previous_results.get(f"{}_{}"):
                #Wrange_lc_Lp,Wrange_rc_Lp,Wrange_core,Wrange_Lp,Wrange_Rp,Frange_Lp,Frange_Rp
            
            #window = [center_window - 500, center_window + 500]
            _d = 0.15
            Wslider_lc = plt.axes([gap_left, bbox_Lp.y0 -_d,  bbox_Lp.width, 0.03])
            Wslider_core = plt.axes([gap_left, bbox_Lp.y0 -_d - 0.05,  bbox_Lp.width, 0.03])
            Wslider_rc = plt.axes([gap_left, bbox_Lp.y0 - _d -0.10,  bbox_Lp.width, 0.03])
            
            Wslider_Lp = plt.axes([bbox_Lp.x0, bbox_Lp.y1*1.01, bbox_Lp.width, 0.03]) 
            Wslider_Rp = plt.axes([bbox_Rp.x0, bbox_Rp.y1*1.01, bbox_Rp.width, 0.03]) 
            
            Fslider_Lp = plt.axes([bbox_Lp.x0 + bbox_Lp.width, bbox_Lp.y0, 0.03, bbox_Lp.height])
            Fslider_Rp = plt.axes([bbox_Rp.x0 + bbox_Rp.width , bbox_Rp.y0, 0.03, bbox_Rp.height]) 
            #Wslider_core_Rp = plt.axes([bbox_Rp.x0, bbox_Rp.y0-0.1, bbox_Rp.width, 0.03]) 
            Wrange_lc_Lp = RangeSlider(Wslider_lc, "left \ncontinium",np.min(X),center_window , valinit=lc_init,color="purple",alpha=0.5) 
            Wrange_rc_Lp = RangeSlider(Wslider_rc, "right \ncontinium",center_window,np.max(X),valinit=lr_init,color="green",alpha=0.2) 
            Wrange_core = RangeSlider(Wslider_core ,"line core",center_window-100,center_window+100,valinit=core_init,color="r",alpha=0.2)
            
            Wrange_Lp = RangeSlider(Wslider_Lp,None,np.min(X),np.max(X),valinit=window)
            Wrange_Rp = RangeSlider(Wslider_Rp,None,np.min(X),np.max(X),valinit=window)
            Frange_Lp = RangeSlider(Fslider_Lp, None, 0, max_local , valinit=[0,q4], orientation='vertical')
            Frange_Rp = RangeSlider(Fslider_Rp,None , -max_local*0.5, max_local , valinit=[-max_local*0.01,max_local*0.5], orientation='vertical')
            
            Wrange_Lp.valtext.set_visible(False)
            Wrange_Rp.valtext.set_visible(False)
            Frange_Lp.valtext.set_visible(False)
            Frange_Rp.valtext.set_visible(False)
            Wrange_lc_Lp.valtext.set_visible(False)
            Wrange_rc_Lp.valtext.set_visible(False)
            Wrange_core.valtext.set_visible(False)
            
            # Use the helper method for the initial plot
            self._window_plot(Lp, Rp, X, Y, objs, center_window, line_name, band, 
                            Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, 
                            Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp)
            
            # Connect sliders to update the plot using the same helper method in the callback
            slider_update = lambda val: self._window_plot(Lp, Rp, X, Y, objs, center_window, line_name, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp)
            
            Wrange_Lp.on_changed(slider_update)
            Frange_Lp.on_changed(slider_update)
            Wrange_Rp.on_changed(slider_update)
            Frange_Rp.on_changed(slider_update)
            Wrange_lc_Lp.on_changed(slider_update)
            Wrange_rc_Lp.on_changed(slider_update)
            Wrange_core.on_changed(slider_update)
            
            save_button = plt.axes([0.4, 0.02, 0.2, 0.04])
            button_save = Button(save_button, 'Save', color='lightgoldenrodyellow', hovercolor='0.975')
            
            button_save.on_clicked(lambda event: self._on_save_button_clicked(event,Lp, Rp, X, Y, objs, center_window, line_name, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp))
            #if name_file in os.listdir(os.getcwd()):
            #   if any((panda_read[["name"]].values == [line_name]).all(axis=1)):
            #      ax_save_button = plt.axes([0.6, 0.02, 0.2, 0.04])
            #     button_remove = Button(ax_save_button, 'remove line', color='lightgoldenrodyellow', hovercolor='0.975')
            #    button_remove.on_clicked(on_remove_line_clicked)
            # Update the plot when slider values change
            #ax_save_button = plt.axes([0.01, 0.95, 0.2, 0.04])
            #button_close = Button(ax_save_button, 'Close', color='lightgoldenrodyellow', hovercolor='0.975')
            #button_close.on_clicked(on_close_all)
            plt.show()
            
            
            return X, Y, window, Lp.get_position()
        
    def windows_analysis(self,band="nir",):
        """_summary_
            overall i will build all around the idea of that the two bands share the same number of pixels because is more easier but
            maybe is more consistent have the idea of this is not always the case.
        """
        
        row = self.pre_define_windows.iloc[8]
        line = row['line_name']
        center_window = np.mean(row["core_range"])
        window = [center_window-500,center_window+500]
        lr_init = row["right_range"]
        lc_init = row["left_range"]
        core_init = row["core_range"]
        #band = "nir"
        X,Y,objs = np.array(self.spectra_dict[band]["wavelength"]), np.array(self.spectra_dict[band]["flux"]),self.spectra_dict[band]["obj"]
        
        if center_window < np.min(X[0]) or center_window > np.max(X[0]):
            return "cant do nothing"
        else:
            w_mask = (X[0]>=min(window)) & (X[0]<=max(window)) #soft coming sooon 
            Y_local = Y[:,w_mask]
            q2 = np.percentile(Y_local, 89)
            q3 = np.percentile(Y_local, 95) 
            q4 = np.percentile(Y_local, 99.97) 
            max_local = np.max(Y_local)
            min_local = np.min(Y_local)
            aspect_ratio = 1.5
            fig = plt.figure(figsize=(20, 15 / 1.5))
            grid = plt.GridSpec(2, 2, width_ratios=[2, 2], height_ratios=[3, 1], hspace=0.4)
            Lp = plt.subplot(grid[0, 0])
            Rp = plt.subplot(grid[0, 1])
            bbox_Lp = Lp.get_position()
            bbox_Rp = Rp.get_position()
            gap_left = bbox_Lp.x0/4 + bbox_Lp.width
            #the val init aqui luego 
            #line -> todo el resto adentro ? lo q seria un doble looop?
            #if self._previous_results.get(f"{}_{}"):
                #Wrange_lc_Lp,Wrange_rc_Lp,Wrange_core,Wrange_Lp,Wrange_Rp,Frange_Lp,Frange_Rp
            
            window = [center_window - 500, center_window + 500]
            _d = 0.15
            Wslider_lc = plt.axes([gap_left, bbox_Lp.y0 -_d,  bbox_Lp.width, 0.03])
            Wslider_core = plt.axes([gap_left, bbox_Lp.y0 -_d - 0.05,  bbox_Lp.width, 0.03])
            Wslider_rc = plt.axes([gap_left, bbox_Lp.y0 - _d -0.10,  bbox_Lp.width, 0.03])
            
            Wslider_Lp = plt.axes([bbox_Lp.x0, bbox_Lp.y1*1.01, bbox_Lp.width, 0.03]) 
            Wslider_Rp = plt.axes([bbox_Rp.x0, bbox_Rp.y1*1.01, bbox_Rp.width, 0.03]) 
            
            Fslider_Lp = plt.axes([bbox_Lp.x0 + bbox_Lp.width, bbox_Lp.y0, 0.03, bbox_Lp.height])
            Fslider_Rp = plt.axes([bbox_Rp.x0 + bbox_Rp.width , bbox_Rp.y0, 0.03, bbox_Rp.height]) 
            #Wslider_core_Rp = plt.axes([bbox_Rp.x0, bbox_Rp.y0-0.1, bbox_Rp.width, 0.03]) 
            Wrange_lc_Lp = RangeSlider(Wslider_lc, "left \ncontinium",np.min(X),center_window , valinit=lc_init,color="purple",alpha=0.5) 
            Wrange_rc_Lp = RangeSlider(Wslider_rc, "right \ncontinium",center_window,np.max(X),valinit=lr_init,color="green",alpha=0.2) 
            Wrange_core = RangeSlider(Wslider_core ,"line core",center_window-100,center_window+100,valinit=core_init,color="r",alpha=0.2)
            
            Wrange_Lp = RangeSlider(Wslider_Lp,None,np.min(X),np.max(X),valinit=window)
            Wrange_Rp = RangeSlider(Wslider_Rp,None,np.min(X),np.max(X),valinit=window)
            Frange_Lp = RangeSlider(Fslider_Lp, None, 0, max_local , valinit=[0,q4], orientation='vertical')
            Frange_Rp = RangeSlider(Fslider_Rp,None , -max_local*0.5, max_local , valinit=[-max_local*0.01,max_local*0.5], orientation='vertical')
            
            Wrange_Lp.valtext.set_visible(False)
            Wrange_Rp.valtext.set_visible(False)
            Frange_Lp.valtext.set_visible(False)
            Frange_Rp.valtext.set_visible(False)
            Wrange_lc_Lp.valtext.set_visible(False)
            Wrange_rc_Lp.valtext.set_visible(False)
            Wrange_core.valtext.set_visible(False)
            
            # Use the helper method for the initial plot
            self._window_plot(Lp, Rp, X, Y, objs, center_window, line, band, 
                            Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, 
                            Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp)
            
            # Connect sliders to update the plot using the same helper method in the callback
            slider_update = lambda val: self._window_plot(Lp, Rp, X, Y, objs, center_window, line, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp)
            
            Wrange_Lp.on_changed(slider_update)
            Frange_Lp.on_changed(slider_update)
            Wrange_Rp.on_changed(slider_update)
            Frange_Rp.on_changed(slider_update)
            Wrange_lc_Lp.on_changed(slider_update)
            Wrange_rc_Lp.on_changed(slider_update)
            Wrange_core.on_changed(slider_update)
            
            save_button = plt.axes([0.4, 0.02, 0.2, 0.04])
            button_save = Button(save_button, 'Save', color='lightgoldenrodyellow', hovercolor='0.975')
            
            button_save.on_clicked(lambda event: self._on_save_button_clicked(event,Lp, Rp, X, Y, objs, center_window, line, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp))
            #if name_file in os.listdir(os.getcwd()):
            #   if any((panda_read[["name"]].values == [line_name]).all(axis=1)):
            #      ax_save_button = plt.axes([0.6, 0.02, 0.2, 0.04])
            #     button_remove = Button(ax_save_button, 'remove line', color='lightgoldenrodyellow', hovercolor='0.975')
            #    button_remove.on_clicked(on_remove_line_clicked)
            # Update the plot when slider values change
            #ax_save_button = plt.axes([0.01, 0.95, 0.2, 0.04])
            #button_close = Button(ax_save_button, 'Close', color='lightgoldenrodyellow', hovercolor='0.975')
            #button_close.on_clicked(on_close_all)
            plt.show()
            
            
            return X, Y, window, Lp.get_position()
    
    def _on_save_button_clicked(self,event,Lp, Rp, X, Y, objs, center_window, line, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp):
        #Add rutine to change the multiple dictionaries in case of been necesary 
        fig, (Lp, Rp) = plt.subplots(1, 2, figsize=(30, 10), gridspec_kw={'width_ratios': [2, 2]})
        self._window_plot(Lp, Rp, X, Y, objs, center_window, line, band, Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp)
        plt.savefig("aja.png", dpi=300, bbox_inches='tight')
        plt.close()
        with open('microlensing.pkl', 'wb') as f:
            print("saved as microlensing.pkl")
            pickle.dump(self.local_list, f)
        
    def _local(self,line_name,imagen,band,x,Wrange_lc_Lp,Wrange_rc_Lp,Wrange_core,Wrange_Lp,Wrange_Rp,Frange_Lp,Frange_Rp):
        result_ = {"line_name":line_name,"band": band,'min_x':np.min(x),
                'max_x': np.max(x),'Wrange_lc_Lp_max': max(Wrange_lc_Lp.val),
                'Wrange_lc_Lp_min': min(Wrange_lc_Lp.val),
                'Wrange_rc_Lp_max': max(Wrange_rc_Lp.val),
                'Wrange_rc_Lp_min': min(Wrange_rc_Lp.val),
                'Wrange_core_max': max(Wrange_core.val),
                'Wrange_core_min': min(Wrange_core.val),
                'Wrange_Lp_max': max(Wrange_Lp.val),
                'Wrange_Lp_min': min(Wrange_Lp.val),
                'Wrange_Rp_max': max(Wrange_Rp.val),
                'Wrange_Rp_min': min(Wrange_Rp.val),
                'Frange_Lp_max': max(Frange_Lp.val),
                'Frange_Lp_min': min(Frange_Lp.val),
                'Frange_Rp_max': max(Frange_Rp.val),
                'Frange_Rp_min': min(Frange_Rp.val)}
        #if self._previous_results:
            #print(imagen)
            #print(self._previous_results[imagen])
         #   prev=self._previous_results.get(f"{imagen}_{band}")
            #print(prev.keys())
            #print([[prev[key],current[key]] for key in current.keys()])
          #  same = all([prev.get(key) == result_.get(key) for key in result_.keys()])
            
            #print(same)
        return result_
    
    
    def _read_previous_results(self, path):
        if isinstance(path, str) and os.path.isfile(path):
            with open(path, 'rb') as f:
                loaded_list = pd.DataFrame(list(pickle.load(f).values()))
                #print(loaded_list)
            return loaded_list
        else:
            return None
        
    #def _compare_dict()
    
    
    
    
    
    def _window_plot(self, Lp, Rp, X, Y, objs, center_window, line, band, 
                     Wrange_lc_Lp, Wrange_rc_Lp, Wrange_core, 
                     Wrange_Lp, Frange_Lp, Wrange_Rp, Frange_Rp):
        """
        Plot the common elements on the provided axes.
        """
        Lp.clear()
        Rp.clear()
        Lp.set_xlim(Wrange_Lp.val)
        Lp.set_ylim(Frange_Lp.val)
        Rp.set_xlim(Wrange_Rp.val)
        Rp.set_ylim(Frange_Rp.val)
        
        Lp.plot(X.T, Y.T, label=objs)
        Lp.legend(framealpha=0, fontsize=12)
        Lp.text(Lp.get_xlim()[0] + 0.01, Lp.get_ylim()[1] * 0.95, f"Window {line} in {band}", fontsize=12)
        Lp.axvline(x=center_window, color='k', linestyle="--")
        Rp.axvline(x=center_window, color='k', linestyle="--")
        
        Lp.fill_betweenx([-100, 100], *Wrange_lc_Lp.val, label='left continuum', color='purple', alpha=0.2)
        Lp.fill_betweenx([-100, 100], *Wrange_rc_Lp.val, label='right continuum', color='g', alpha=0.2)
        Lp.fill_betweenx([-100, 100], *Wrange_core.val, label='Core', color='r', alpha=0.2)
        Rp.fill_betweenx([-100, 100], *Wrange_core.val, label='Core', color='r', alpha=0.2)
        Rp.hlines(y=[0, 0], xmin=np.min(X), xmax=np.max(X), colors='k', linestyles='dashed', zorder=3)
        self.local_list = {}
        for i, (x, y, key) in enumerate(zip(X, Y, objs)):
            mask_lc = (x >= min(Wrange_lc_Lp.val)) & (x <= max(Wrange_lc_Lp.val))
            mask_rc = (x >= min(Wrange_rc_Lp.val)) & (x <= max(Wrange_rc_Lp.val))
            x_combined = np.concatenate((x[mask_lc], x[mask_rc]))
            y_combined = np.concatenate((y[mask_lc], y[mask_rc]))
            slope_fit, intercept_fit = linear_model(x_combined, y_combined)
            y_fit = linear_func(x_combined, slope_fit, intercept_fit)
            Lp.plot(x_combined, y_fit, label=f'Fitted Linear Function for {key}', color='k', linestyle="--")
            y_without_cont = y - linear_func(x, slope_fit, intercept_fit)
            Rp.plot(x, y_without_cont, label=key, alpha=0.5)
            self.local_list[f"{key}_{band}"] = self._local(line,key,band,x,Wrange_lc_Lp,Wrange_rc_Lp,Wrange_core,Wrange_Lp,Wrange_Rp,Frange_Lp,Frange_Rp)
            #.append(self._local(line,key,band,x,Wrange_lc_Lp,Wrange_rc_Lp,Wrange_core,Wrange_Lp,Wrange_Rp,Frange_Lp,Frange_Rp))
            #self.local_ = 
            #         # Calculate the area under the linear function between Barrier 1 and Barrier 4
            #area, _ = quad(linear_fit, x_barrier1, x_barrier4, args=(slope_fit, intercept_fit))
            #y_curve = y_noisy - linear_fit(x, slope_fit, intercept_fit)
            #Y_curve = Y[key] - linear_fit(X, slope_fit, intercept_fit)
            #suma = np.sum(y_curve[(x_barrier5 <= x) & (x_barrier6 >= x)])

        Rp.legend(framealpha=0, fontsize=12)
        Lp.tick_params(which="both", length=10, width=2, labelsize=20)
        Rp.tick_params(which="both", length=10, width=2, labelsize=20)
        Lp.set_ylabel(r"$f_\lambda\,(\mathrm{erg\,s^{-1}\,cm^{-2}\,\AA^{-1}})$", fontsize=20)
        Lp.set_xlabel(r'$\rm Rest \ Wavelength$ ($\rm \AA$)', fontsize=20)
        Rp.set_ylabel(r"$f_\lambda\,(\mathrm{erg\,s^{-1}\,cm^{-2}\,\AA^{-1}})$", fontsize=20)
        Rp.set_xlabel(r'$\rm Rest \ Wavelength$ ($\rm \AA$)', fontsize=20)
        plt.draw()
        
        


import pandas as pd
import numpy as np
from copy import deepcopy
import re
import os
#from .modeling import *
import pickle
#import h5py

module_dir = os.path.dirname(os.path.abspath(__file__))


def write_list_to_file(file_path, data_lines):
    """
    Write the content of a list to a file.
    
    Args:
        file_path (str): Path to the file to save.
        data_lines (list of str): Lines to write to the file.
    """
    try:
        with open(file_path, "w") as file:
            file.writelines(data_lines)
        print(f"File successfully written to: {file_path}")
    except Exception as e:
        print(f"An error occurred: {e}")

def clean_number(txt):
    if txt == "-":
        return np.nan
    if "/" in str(txt):
        return np.nan
    else:
        return str(txt).translate(str.maketrans('', '', '()~?'))
    
    
def bands_that_can_be_modeled(L:pd.DataFrame,print_bands=False,keyword="IS"):
    "given a panda data frame for and unique object choose the best band for our experiment e.g. I/F814/i"
    
    try:
        L=L.loc[L.index.get_level_values('type')=="ima"]
    except:
        L=L.loc[L[keyword]=="ima"]
    bands = [i for i in L.columns if "band" in i if sum(~np.isnan(L[i].replace("-",np.nan).astype(float)))==len(L[L[keyword]=="ima"])]
    if print_bands:
        print(f"avaliable bans {bands}")
    if len(bands)>0:
        for i in ["band_i",'band_I',"band_F814W"]:
            if i in bands:
                bands = [i]+list(set(bands) - set([i]))
    else:
        bands = ["-"]
    return bands
#This function will require take in consideration the cases in where the number of images dosent mach the positions
def can_be_modeled(L:pd.DataFrame,keywords=["IS","lens"],to_pandas=True):
    "cbm:can be modeled ? given a panda data frame for and unique object "
    system_name = L.name.drop_duplicates().values[0]
    bands=bands_that_can_be_modeled(L)
    #maybe here will be require change the way is added the images
    try:
        imagenes=L[L[keywords[0]]=="ima"][["RA","DEC","dDEC","dRA"]].astype(float) #informacion de las imagenes
    except:
        imagenes=[] #informacion de las imagenes
    n_images = len(imagenes)
    dif=L[L[keywords[0]]==keywords[1]][[i for i in L[L[keywords[0]]==keywords[1]].columns if i not in ["RA","DEC","dDEC","dRA"]]]
    problems=[]
    cbmd=False
    conditions = {
    n_images == 0: "0 images",
    n_images == 1: "ring",
    n_images > 5: "more than 5 images",
    n_images == 3: "it has 3 images",
    bands == ["-"]: "no band information",
    len(dif) == 0: "no information about the galaxy"
}
    for condition, message in conditions.items():
        if condition:
            problems.append(message)
    cbmd = len(problems) == 0
    # If there are no problems, set problems to ["-"]
    if cbmd:
        problems = ["-"]
        if to_pandas==True:
            return [cbmd,problems,bands] 
    #np.tile(np.array([cbmd,*problems,bands]), (len(L),1 ))
        return [cbmd,system_name,problems]
    else:
        return [cbmd,problems,bands] 

def columns_to_float(system,band):
    dif= system[["RA","DEC","dDEC","dRA"]][system['IS']=="lens"][["RA","DEC","dDEC","dRA"]].astype(float)
    columns_to_float = ['RA', 'DEC', 'dDEC', 'dRA', band,band.replace("band","error")]
    images = system[system['IS'] == 'ima'][["component"] + ['RA', 'DEC', 'dDEC', 'dRA', band,band.replace("band","error")]].astype({col: float for col in columns_to_float})
    return images,dif

def is_know_redshift(Lenses:pd.DataFrame):
    "Given a data frame of lens system look for redshift in the system in the current table and replace - for 0.5 or 2.0"
    if "z_l" not in Lenses.columns:
        Lenses["z_l"] = ["-"] * len(Lenses)
    if "z_s" not in Lenses.columns:
        Lenses["z_s"] = ["-"] * len(Lenses)
    Lenses["zl_known"] = [bool(zl != "-" and ~np.isnan(float(clean_number(zl)))) for zl in Lenses["z_l"].values]
    Lenses["zs_known"] = [bool(zs != "-" and ~np.isnan(float(clean_number(zs)))) for zs in Lenses["z_s"].values]
    Lenses["old_z_s"] = [clean_number(zs) if known else np.nan for known,zs in Lenses[["zs_known","z_s"]].values]
    Lenses["old_z_l"] = [clean_number(zl) if known else np.nan for known,zl in Lenses[["zl_known","z_l"]].values]
    Lenses["z_s"] = [clean_number(zs) if known else 2.0 for known,zs in Lenses[["zs_known","z_s"]].values]
    Lenses["z_l"] = [clean_number(zl) if known else 0.5 for known,zl in Lenses[["zl_known","z_l"]].values]
    return Lenses

def model_available(imagenes:float,dif:float):
    #all use SIE
    Models=[]
    if imagenes==1:
        Models=["SIS"]
    if imagenes==2:
        Models=["SIS","SIE","SIS+shear"]
    if imagenes==3:
        Models=["SIS","SIE","SIS+shear"]
    if imagenes==4:
        Models=["SIS","SIE","SIE+shear","POW","POW+shear"]
        if dif>1:
            Models.append("SIE-2G")
    if imagenes>=5 and dif>1:
        Models.append("SIE-2G")
    return Models

def image_whit_photometry(L:pd.DataFrame,verbose=False):
    system = deepcopy(L)
    img_all = system[system["IS"]=="ima"]["component"].values
    img_can_be_moleded = system[system["IS"]=="ima"].dropna(subset=[i for i in system.columns if "band" in i], how='all')
    if len(img_can_be_moleded)==0:
        if verbose:
            print(f"No image for system {system.name.values[0]}")
        return img_can_be_moleded
    
    system["total_images"] = np.repeat(len(img_all),len(system))
    system_clean = pd.concat([img_can_be_moleded, system[system["IS"]!="ima"]], ignore_index=True)
    image_cant_be_modeled =list(set(system[system["IS"]=="ima"]["component"].values)-set(img_can_be_moleded["component"].values))
    system_clean["model_images"] = np.repeat(len(img_can_be_moleded),len(system_clean))
    if len(image_cant_be_modeled)==0:
        image_cant_be_modeled = ["-"]
    system_clean["image_cant_be_modeled"] = len(system_clean)*image_cant_be_modeled
    system_clean = system_clean.dropna(subset=["RA","DEC"])
    system_clean = system_clean.drop_duplicates(subset=['RA',"DEC"])
    system_clean["total_lens"] = np.repeat(len(system_clean[system_clean["IS"]=="lens"]),len(system_clean))
    system_clean["model_available"] = len(system_clean)*[model_available(system_clean.model_images.values[0],system_clean.total_lens.values[0])]
    cbmd,problems,bands = can_be_modeled(system_clean)
    system_clean["can_be_modeled"] =len(system_clean) * [cbmd]
    system_clean["issues"] =len(system_clean) * problems
    system_clean["available_bands"] =len(system_clean) * [bands]
    return system_clean

def pandas_to_model(Lenses:pd.DataFrame,verbose=False,only_astrometry=False):
    if "IS" not in Lenses.columns:
        if verbose:
            print("IS not in columns")
        Lenses["can_be_modeled"] = [False]*len(Lenses)
        return Lenses
    if "ima" not in Lenses.IS.values:
        return Lenses
    if not any(["G" in str(i) for i in Lenses["component"].values]):
        if verbose:
            print("not G in the system")
        Lenses["can_be_modeled"] = [False]*len(Lenses)
        return Lenses
    Lenses_f = pd.concat([image_whit_photometry(Lenses[Lenses["name"]==name]) for name in Lenses.name.drop_duplicates()],ignore_index=True)
    if len(Lenses_f)==0:
        Lenses["can_be_modeled"] = [False]*len(Lenses)
        return Lenses
    Lenses_f = is_know_redshift(deepcopy(Lenses_f))
    if 'photometric_system' not in Lenses_f.columns:
        Lenses_f['photometric_system'] = None
    if only_astrometry:
        Lenses_f = Lenses_f[['name', 'z_l', 'z_s', 'ra', 'dec', 'component', 'IS', 'RA', 'dRA','DEC', 'dDEC','Bibcode', 'file', 'year', 'Bibcode_zl', 'Bibcode_zs','known_names', 'can_be_modeled', 'total_images', 'model_images','image_cant_be_modeled', 'total_lens', 'model_available', 'issues', 'zl_known', 'zs_known', 'old_z_s', 'old_z_l','photometric_system','Telescope', 'instrument']]
    return Lenses_f



def get_paths_before(target_path):
    # Get the absolute path of the target
    target_abs_path = os.path.abspath(target_path)
    
    # Split the path into components
    path_components = target_abs_path.split(os.sep)
    
    # Remove the last component (the target)
    path_components.pop()
    
    # Reconstruct the paths before the target
    paths_before = []
    for i in range(len(path_components)):
        paths_before.append(os.sep.join(path_components[:i+1]))
    
    return paths_before


def make_dictionary(list,prefix="p"):
    return {f"{prefix}[{i}]":float(list[i]) for i in range(len(list))}

def get_numeric_values(text):
    numeric_values=re.findall(r'[-+]?\d*\.\d+|\d+|\(?\s*[-+]?\d*\.?\d*e[+-]?\d*\s*\)?', text)
    return numeric_values

def get_images_result(list_lensmodel):
    images_result = {"ra_imput":[],"dec_imput":[],"radec_error":[],"flux_imput":[],"flux_error":[],"time_delay_imput":[],"time_delay_error":[],"ra_output":[],"dec_output":[],"flux_output":[],"time_delay_output":[]}
    for j,imagen in enumerate(list_lensmodel):
        if imagen=='\n':
            continue
        imagen = imagen.replace("\n","").split("->")
        numeric_values =  [i.replace(")","").replace("(","") for i in  get_numeric_values(imagen[0])]+[float(i) for i in get_numeric_values(imagen[1])]
        for i,key in enumerate(images_result.keys()): images_result[key].append(float(numeric_values[i]))
    return images_result

def get_kappa_gamma(list_of_lines):
    kappa_gamma = {"index":[],"x":[],"y":[],"kappa":[],"gamma":[],"theta":[]}
    for j in range(2,len(list_of_lines)):
        linea = [i for i in list_of_lines[j].replace("\n","").split(" ") if i!=""]
        for i,key in enumerate(kappa_gamma.keys()): kappa_gamma[key].append(float(linea[i]))
    kappa_gamma = {key: np.array(value) for key, value in kappa_gamma.items()}
    kappa_gamma["magnification"] =((1-kappa_gamma['kappa'])**2 - kappa_gamma['gamma']**2)**-1
    return {"kappa_gamma":kappa_gamma}

def get_result_lensmodel(list_of_lines):
    possible_keys=['LENS PARMS','SOURCE PARMS','CHISQ','Source','images','Extra model images']
    where_appear= np.array([[pk,i] for pk in possible_keys for i,line in enumerate(list_of_lines) if pk == line.split(":")[0]])
    where_appear_final = np.vstack((where_appear.T,np.hstack((where_appear[1:,1],-1)))).T
    result = {str(k):{} for k in where_appear_final[:,0]}
    for k,f_i,u_i in where_appear_final:
        #print(k,f_i,u_i)
        local_list = list_of_lines[int(f_i):int(u_i)]
        if k=='LENS PARMS':
        #local_list = list_of_lines[int(f_i):int(u_i)]
            for i,line in enumerate(local_list):
                if "alpha" in line:
                    result["LENS PARMS"][f"alpha{i}"] = make_dictionary([i.replace("\n","") for i in line.split(" ") if (len(i)>0 and i!='alpha')])
        if k=="SOURCE PARMS":
            for i,line in enumerate(local_list):
                if "ptsrc" in line:
                        result["SOURCE PARMS"][f"ptsrc{i}"] = make_dictionary([i.replace("\n","") for i in line.split(" ") if (len(i)>0 and i!='ptsrc')],prefix="s")
        if k=="CHISQ":
            chi_values,chi_columns = list_of_lines[int(f_i)].split("#")
            result["CHISQ"]= {v: float(k) for k, v in zip(chi_values.replace("CHISQ:","").split(" ")[1:-1],chi_columns.replace("\n","").split(" ")[1:])}
            result["Source"] = {local_list[i].replace("\n","").split(":")[0].replace(" ",""):local_list[i].replace("\n","").split(":")[1]  for i in range(3,len(local_list[1:])+1)} #only works for one source 
        if k=="images":
            result["images"] = get_images_result(local_list[1:])
    return result
def get_critical_caustic(list_of_lines):
    return {"critical_caustic":{key:np.array([[i for i in i.split(" ")[1:] if len(i)>0] for i in [i.replace("\n","") for i in list_of_lines][1:] if len(i)>0],dtype="object").astype(float)[::,n] for n,key  in enumerate(["x","y","u","v"])}}

def get_grid(list_of_lines):
    return {"grid":{key:np.array([[i for i in i.split(" ") if len(i)>0] for i in [i.replace("\n","") for i in list_of_lines][1:] if len(i)>0],dtype="object").astype(float)[::,n] for n,key  in enumerate(["x","y","u","v"])}}

def get_RE(list_of_lines):
   re = {'RE':float(list_of_lines[0].split(" ")[0])}
   return re

def full_modeling_result(modeling_path,model_setup,remove_files=True):
    #where can add the name of the system xd
    model_result = {}
    ###
    if "final_step.start" in os.listdir(modeling_path):
            model_result["best_model_file"] = open(os.path.join(modeling_path,"final_step.start"), "r").readlines()
    if f"{modeling_path.split('/')[-1]}.dat" in os.listdir(modeling_path):
        model_result["model_setup_file"] = open(os.path.join(modeling_path,f"{modeling_path.split('/')[-1]}.dat"), "r").readlines()
    if f"{model_setup['name_run']}.dat" in os.listdir(modeling_path):
        model_result["modeling_file"] = open(os.path.join(modeling_path,f"{model_setup['name_run']}.dat"), "r").readlines()
    if "final_step.dat" in os.listdir(modeling_path):
        model_result["final_step_file"] = open(os.path.join(modeling_path,"final_step.dat"), "r").readlines()
    if "final_step.best" in os.listdir(modeling_path):
        model_result["final_step_file"] = open(os.path.join(modeling_path,"final_step.best"), "r").readlines()
    ####
    for step in ["step_1","step_2","final_step","kappa_gamma","critical_caustic","grid","RE"]: #only step+"kappa_gamma","critical_caustic","grid"   
        file_name = f"{step}.dat"
        if "step" in step:
            func = get_result_lensmodel
            if f"{step}.best" in os.listdir(modeling_path):
                file_name = f"{step}.best"
        elif step=="kappa_gamma":
            func = get_kappa_gamma
        elif step=="critical_caustic":
            func = get_critical_caustic
        elif step=="grid":
            func = get_grid
        elif step=="RE":
            func =get_RE
        if file_name not in  os.listdir(modeling_path):
            continue
        file = open(os.path.join(modeling_path,file_name), "r")
        list_of_lines = file.readlines()
        model_result[step] = func(list_of_lines)
        if remove_files:
            [os.remove(os.path.join(modeling_path,i) )for i in os.listdir(modeling_path) if step in i]
    model_result["model_setup"] = {"model_setup":model_setup}
    return model_result


def compare_dicts(dict1, dict2):
    # Check if both dictionaries have the same keys
    if set(dict1.keys()) != set(dict2.keys()):
        return False
    
    # Check if the values for each key are the same
    false_key=[]
    for key in dict1.keys():
        if np.all(dict1[key] != dict2[key]):
            false_key.append(key)
    if len(false_key)>0:
        #print(false_key)
        return False
    return True




def read_pickle(file_path):
    with open(file_path, 'rb') as handle:
        data = pickle.load(handle)
    return data


def write_pickle(data,file_path):
    #maybe add and istance dic?
    with open(file_path, 'wb') as handle:
        #print(filename.replace("hdf5","pickle"))
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

class ModelNotFoundError(Exception):
    def __init__(self, model, available_models):
        message = f"The model '{model}' is not in the lensmodel keys list: {available_models}"
        super().__init__(message)






def format_scientific(value, precision=3):
    """
    Format a value in scientific notation with a specified number of decimal places.
    
    Parameters:
        value (float): The value to be formatted.
        precision (int): The number of decimal places for the coefficient.

    Returns:
        str: The formatted value in scientific notation.
    """
    coefficient, exponent = "{:.{precision}e}".format(value, precision=precision).split('e')
    return f"{float(coefficient):.{precision}f} × 10^{int(exponent)}"



def models_disp(imagenes:list,dif:list):
    #all use SIE
    Models=[]
    if len(imagenes)==1:
        Models=["SIS"]
    if len(imagenes)==2:
        Models=["SIS","SIE","SIS+shear"]
    if len(imagenes)==3:
        Models=["SIS","SIE","SIS+shear"]
    if len(imagenes)==4:
        Models=["SIE","SIE+shear"]
        if len(dif)>1:
            Models.append("SIE-2G")
    if len(imagenes)>=5 and len(dif)>1:
        Models.append("SIE-2G")
    return Models
def writte_lensmodel_file():
    return


import multiprocessing
import pandas as pd
import numpy as np
from .lensmodel_wrapper import run_lensmodel
from .utils import get_RE,write_pickle,get_kappa_gamma,get_kappa_gamma#,get_critical_caustic,get_grid,clean_number
import os 

def make_pandas_from_mcmc(path):
    # Initialize lists to hold data
    f = open(f"{path}/mcmc.dat","r")
    file_content=f.readlines()
    lens_parms = []
    source_parms = []
    chisq_values = []

    # Iterate over the lines and extract data
    for i, line in enumerate(file_content):
        if "LENS PARMS" in line:
            lens_values = list(map(str, file_content[i+1].strip().split()[0:]))
            lens_parms.append(lens_values)
        elif "SOURCE PARMS" in line:
            source_values = list(map(str, file_content[i+1].strip().split()[0:]))
            source_parms.append(source_values)
        elif "CHISQ" in line:
            chisq_value = float(line.split()[1])
            chisq_values.append(chisq_value)

    # Combine into a DataFrame
    data = {
        'Lens_mass': [params[0] for params in lens_parms],
        'p[1]': [params[1] for params in lens_parms],
        'p[2]': [params[2] for params in lens_parms],
        'p[3]': [params[3] for params in lens_parms],
        'p[4]': [params[4] for params in lens_parms],
        'p[5]': [params[5] for params in lens_parms],
        'p[6]': [params[6] for params in lens_parms],
        'p[7]': [params[7] for params in lens_parms],
        'p[8]': [params[8] for params in lens_parms],
        'p[9]': [params[9] for params in lens_parms],
        'p[10]': [params[10] for params in lens_parms],
        'Source': [params[0] for params in source_parms],
        'Source_Param1': [params[1] for params in source_parms],
        'Source_Param2': [params[2] for params in source_parms],
        'Source_Param3': [params[3] for params in source_parms],
        'Source_Param4': [params[4] for params in source_parms],
        'Source_Param5': [params[5] for params in source_parms],
        'Source_Param6': [params[6] for params in source_parms],
        'Source_Param7': [params[6] for params in source_parms],
        'CHISQ': chisq_values
    }

    df = pd.DataFrame(data)
    return df
def modify_alpha_row(path,data, new_values):
    """
    Modify the values in the 'alpha' row of the data list.

    Parameters:
    data (list): The list containing the rows of data as strings.
    new_values (list): A list of new values to replace the current values in the 'alpha' row.

    Returns:
    list: The modified list with updated 'alpha' row values.
    """
    # Find the row containing the 'alpha' keyword
    alpha_row_index = next((i for i, row in enumerate(data) if 'alpha' in row), None)

    if alpha_row_index is not None:
        # Split the alpha row into components
        components = data[alpha_row_index].split()

        # Replace the numerical values with new values
        # Assuming the 'alpha' keyword is the first component
        components[1:len(new_values)+1] = new_values

        # Recombine the components into a single string
        data[alpha_row_index] = ' '.join(components) + '\n'
    with open(f'{path}/start_re_kg.dat', 'w') as file:
        file.writelines(data)

def make_mcmc_lensmodel(system_class,model_to_mcmc,path,rewrite=False):
    model,system_info = system_class.get_necessary_to_mcmc(model_to_mcmc)
    if not rewrite:
        if "mcmc" in model.keys():
            print(f"MCMC alredy done for the model {model_to_mcmc} and rewritte false")
            return system_class
    system_file,start_mcmc = model["model_setup_file"],model["best_model_file"]
    z_l,z_s,name = system_info["zl"],system_info["zs"],system_info["name"]
    print("Starting MCMC")
    with open(f'{path}/{name}.dat', 'w') as file:
        file.writelines(system_file)
    with open(f'{path}/start_mcmc.dat', 'w') as file:
        file.writelines(start_mcmc)
    with open(f'{path}/domcmc.dat', 'w') as file:
        mcmc = [    "#MCMC\n",
                    "set omitcore = 0.05\n",
                    f"data {path}/{name}.dat #filewhitdata\n",
                    "set omega = 0.3\n",
                    "set lambda = 0.7\n",
                    "set hval = 0.7\n",
                    "set hvale = 1000 # Uncertainty in H_0 in units of 100 km/s/Mpc\n",
                    "set chimode = 0 #sourceplane\n",
                    f"set zlens = {z_l} # lens redshift\n",
                    f"fset zsrc = {z_s} # source redshift\n",
                    "set checkparity = 0 # don't worry about parities\n",
                    "set gridflag = 0 # don't need the tiling\n",
                    "#set restart = 2\n",
                    f"setlens {path}/start_mcmc.dat\n",
                    "MCMCset 2 Nchain 10 maxsteps 100000\n",
                    f"MCMCrun {path}/mcmc\n","quit\n"]
        file.writelines(mcmc)
    run_lensmodel(path,"domcmc")
    result_pandas=make_pandas_from_mcmc(path)
    with open(f'{path}/do_re_kg.dat', 'w') as file:
        main = [
                    "set omitcore = 0.05\n",
                    f"data {path}/{name}.dat #filewhitdata\n",
                    "set omega = 0.3\n",
                    "set lambda = 0.7\n",
                    "set hval = 0.7\n",
                    "set hvale = 1000 # Uncertainty in H_0 in units of 100 km/s/Mpc\n",
                    "set chimode = 0 #sourceplane\n",
                    f"set zlens = {z_l} # lens redshift\n",
                    f"fset zsrc = {z_s} # source redshift\n",
                    "set checkparity = 0 # don't worry about parities\n",
                    "set gridflag = 0 # don't need the tiling\n",
                    "#set restart = 2\n",
                    f"setlens {path}/start_re_kg.dat\n",
                    f"calcRein 10 {path}/RE.dat\n",
                    f"kapgam 3 {path}/kappa_gamma.dat\n",
    "quit\n"]
        file.writelines(main)
    RE = []
    data_list = []
    for new_values in result_pandas[[i for i in result_pandas.columns.values if "p[" in i]].values:
        #print(new_values)
        modify_alpha_row(path,start_mcmc, new_values)
        run_lensmodel(path,"do_re_kg")
        RE.append( get_RE(open(os.path.join(path,"RE.dat"), "r").readlines())["RE"])
        data_list.append(get_kappa_gamma(open(os.path.join(path,"kappa_gamma.dat"), "r").readlines())["kappa_gamma"])
    result_pandas["RE"] = RE 
    model.update({"mcmc":{"system_info":system_info,"mcmc_chain":result_pandas,"kappa_gamma_chain": pd.concat([pd.DataFrame(data) for data in data_list], ignore_index=True)}})
    system_class.lensmodel_system[model_to_mcmc] = model
    write_pickle(system_class.lensmodel_system,f"{path}/model_result.pickle")
    [os.remove(os.path.join(path,i) )for i in os.listdir(path) if "model_result.pickle" not in i]
    print("MCMC ENDED")
    return system_class

# def make_mcmc_lensmodel_file(path,model_to_do_mcmc,system,max_separation,converge,rewrite=False):
#     """given the format of my code i prefer use list that contains the init and start file the start file is the best values file"""
#     if rewrite==False and os.path.isfile(f"{path}/mcmc.pickle"):
#         print("MCMC alredy done")
#         return
    
#     name = path.split("/")[-1]
#     model_name = model_to_do_mcmc["model_name"]
#     system_file,start_mcmc = model_to_do_mcmc["model_setup_file"],model_to_do_mcmc["best_model_file"]
#     z_l,z_s = float(clean_number(system.z_l.values[0])),float(clean_number(system.z_s.values[0]))
#     print("Starting mcmc")
#     with open(f'{path}/{name}.dat', 'w') as file:
#         file.writelines(system_file)
#     with open(f'{path}/start_mcmc.dat', 'w') as file:
#         file.writelines(start_mcmc)
#     with open(f'{path}/domcmc.dat', 'w') as file:
#         mcmc = [
#                     "#MCMC\n",
#                     "set omitcore = 0.05\n",
#                     f"data {path}/{name}.dat #filewhitdata\n",
#                     "set omega = 0.3\n",
#                     "set lambda = 0.7\n",
#                     "set hval = 0.7\n",
#                     "set hvale = 1000 # Uncertainty in H_0 in units of 100 km/s/Mpc\n",
#                     "set chimode = 0 #sourceplane\n",
#                     f"set zlens = {z_l} # lens redshift\n",
#                     f"fset zsrc = {z_s} # source redshift\n",
#                     "set checkparity = 0 # don't worry about parities\n",
#                     "set gridflag = 0 # don't need the tiling\n",
#                     "#set restart = 2\n",
#                     f"setlens {path}/start_mcmc.dat\n",
#                     "MCMCset 2 Nchain 10 maxsteps 100000\n",
#                     f"MCMCrun {path}/mcmc\n",
                    
#     "quit\n"
# ]
    
#         file.writelines(mcmc)
    
#     run_lensmodel(path,"domcmc")
#     result_pandas=make_pandas_from_mcmc(path)
    
#     with open(f'{path}/do_re_kg.dat', 'w') as file:
#         main = [
#                     "set omitcore = 0.05\n",
#                     f"data {path}/{name}.dat #filewhitdata\n",
#                     "set omega = 0.3\n",
#                     "set lambda = 0.7\n",
#                     "set hval = 0.7\n",
#                     "set hvale = 1000 # Uncertainty in H_0 in units of 100 km/s/Mpc\n",
#                     "set chimode = 0 #sourceplane\n",
#                     f"set zlens = {z_l} # lens redshift\n",
#                     f"fset zsrc = {z_s} # source redshift\n",
#                     "set checkparity = 0 # don't worry about parities\n",
#                     "set gridflag = 0 # don't need the tiling\n",
#                     "#set restart = 2\n",
#                     f"setlens {path}/start_re_kg.dat\n",
#                     f"calcRein 10 {path}/RE.dat\n",
#                     f"kapgam 3 {path}/kappa_gamma.dat\n",
#     "quit\n"]
#         file.writelines(main)
#     RE = []
#     data_list = []
#     for new_values in result_pandas[[i for i in result_pandas.columns.values if "p[" in i]].values:
#         #print(new_values)
#         modify_alpha_row(path,start_mcmc, new_values)
#         run_lensmodel(path,"do_re_kg")
#         RE.append( get_RE(open(os.path.join(path,"RE.dat"), "r").readlines())["RE"])
#         data_list.append(get_kappa_gamma(open(os.path.join(path,"kappa_gamma.dat"), "r").readlines())["kappa_gamma"])
#     result_pandas["RE"] = RE 
#     model_result = {"mcmc_chain":result_pandas,"system":system,"kappa_gamma_chain": pd.concat([pd.DataFrame(data) for data in data_list], ignore_index=True)}
#     model_result.update(model_to_do_mcmc)
#     #model_result[i for i in model_to_do_mcmc.keys()] = {i:model_to_do_mcmc[i] for i in model_to_do_mcmc.keys()}
#     model_result["max_separation"] = max_separation
#     model_result["converge"] = converge
#     for i in os.listdir(path):
#         #i can remove this some day
#         if i =="final_result":
#             continue
#         if i.endswith(('.pickle', '.hdf5')):
#             continue
#         os.remove(os.path.join(path,i))
#     main_result = {}
#     main_result[model_name+"_"+str(max_separation)+"_"+str(converge)] = model_result
#     write_pickle(main_result,f"{path}/mcmc.pickle")
#     print(f"{path}/mcmc.pickle")
    
    
    


import os
import stat
import subprocess

module_dir = os.path.dirname(os.path.abspath(__file__))

def run_lensmodel(modeling_path,run_name):
    try:
        path_to_run = os.path.join(modeling_path,f"{run_name}.dat")
        where_we_are = os.getcwd()
        path_to_lensmodel = f"{module_dir}/lensmodel/lensmodel"
        #print(path_to_lensmodel)
        st = os.stat(path_to_lensmodel)
        # Get permissions and make the file executable
        os.chmod(path_to_lensmodel, st.st_mode | stat.S_IEXEC)
        # Change the current working directory to work_pad
        os.chdir(os.path.join(module_dir,f"{module_dir}/lensmodel"))
        with open(path_to_run, 'r') as f:
            proc = subprocess.Popen([path_to_lensmodel], stdin = f, stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
            proc.wait()
        os.chdir(where_we_are)
    except KeyboardInterrupt:
        os.chdir(where_we_are)
    #just remove the files in lensmodel
    [os.remove(os.path.join(f"{module_dir}/lensmodel",i) )for i in os.listdir(f"{module_dir}/lensmodel") if "lensmodel" not in i]
        #print("Execution stopped by the user.")



import matplotlib.pyplot as plt

plt.rcParams.update({
    
    'lines.linewidth': 3,
    'lines.color': '#929591',        # Default color for lines
    'axes.labelsize': 20,            # Label font size
    'axes.titlesize': 20,            # Title font size
    'axes.labelcolor': 'black',      # Color of axis labels
    'xtick.labelsize': 15,           # X tick font size
    'ytick.labelsize': 15,           # Y tick font size
    'xtick.direction': 'in',         # X tick direction
    'ytick.direction': 'in',         # Y tick direction
    'legend.fontsize': 22,           # Legend font size
    'legend.frameon': False,         # No frame around legend
    'legend.loc': 'upper center',    # Default legend position
    'figure.figsize': (35, 15)       # Default figure size
})







def clean_name(L):
    "remove leters from the name, caracteres, hopefully we can obtain compare those with previous results"
    if not isinstance(L,str):
        l=[]
        for ii in L:
            for i in [chr(i) for i in range(ord('A'),ord('Z')+1)]:
                    ii=ii.replace(i,"").replace("†","").replace("−","-").replace("*","")
            l.append(ii)
        L=l
    else:
        for i in [chr(i) for i in range(ord('A'),ord('Z')+1)]:
            L=L.replace(i,"").replace("†","").replace("−","-").replace("*","")
    return L




### all of this should be here 
def make_dictionary(list,prefix="p"):
    return {f"{prefix}[{i}]":float(list[i]) for i in range(len(list))}

def get_numeric_values(text):
    numeric_values=re.findall(r'[-+]?\d*\.\d+|\d+|\(?\s*[-+]?\d*\.?\d*e[+-]?\d*\s*\)?', text)
    return numeric_values

def get_images_result(list_lensmodel):
    images_result = {"ra_imput":[],"dec_imput":[],"radec_error":[],"flux_imput":[],"flux_error":[],"time_delay_imput":[],"time_delay_error":[],"ra_output":[],"dec_output":[],"flux_output":[],"time_delay_output":[]}
    for j,imagen in enumerate(list_lensmodel):
        imagen = imagen.replace("\n","").split("->")
        numeric_values =  [i.replace(")","").replace("(","") for i in  get_numeric_values(imagen[0])]+[float(i) for i in get_numeric_values(imagen[1])]
        for i,key in enumerate(images_result.keys()): images_result[key].append(float(numeric_values[i]))
    return images_result

def get_kappa_gamma(list_of_lines):
    kappa_gamma = {"index":[],"x":[],"y":[],"kappa":[],"gamma":[],"theta":[]}
    for j in range(2,len(list_of_lines)):
        linea = [i for i in list_of_lines[j].replace("\n","").split(" ") if i!=""]
        for i,key in enumerate(kappa_gamma.keys()): kappa_gamma[key].append(float(linea[i]))
    kappa_gamma = {key: np.array(value) for key, value in kappa_gamma.items()}
    kappa_gamma["magnification"] =((1-kappa_gamma['kappa'])**2 - kappa_gamma['gamma']**2)**-1
    return {"kappa_gamma":kappa_gamma}

def get_result_lensmodel(list_of_lines):
    result = {}
    for n,line in enumerate(list_of_lines):
        if "LENS PARMS" in line:
            #what happend if i have 2 galaxies ?
            result["LENS PARMS"] ={}
            for i in range(1,10):
                if "alpha" in list_of_lines[n+i]:
                    result["LENS PARMS"][f"alpha{i}"] = make_dictionary([i.replace("\n","") for i in list_of_lines[n+1].split(" ") if (len(i)>0 and i!='alpha')])
        if "SOURCE PARMS" in line:
            result["SOURCE PARMS"] ={}
            for i in range(1,10):
                if "ptsrc" in list_of_lines[n+i]:
                    result["SOURCE PARMS"][f"ptsrc{i}"] = make_dictionary([i.replace("\n","") for i in list_of_lines[n+1].split(" ") if (len(i)>0 and i!='ptsrc')],prefix="s")
        if "CHISQ" in line:
            chi_values,chi_columns = list_of_lines[n].split("#")
            result["CHISQ"]= {v: float(k) for k, v in zip(chi_values.replace("CHISQ:","").split(" ")[1:-1],chi_columns.replace("\n","").split(" ")[1:])}
        if "Source" in line:
            #should be updated for cases with more source
            result["Source"] = {i[0].replace(" ",""):i[1] for i in [list_of_lines[n+i].replace("\n","").split(":") for i in range(1,6)]}
        if "images" in line:
            result["images"] = get_images_result(list_of_lines[n:][1:-2])
    return result

def get_critical_caustic(list_of_lines):
    return {"critical_caustic":{key:np.array([[i for i in i.split(" ")[1:] if len(i)>0] for i in [i.replace("\n","") for i in list_of_lines][1:] if len(i)>0],dtype="object").astype(float)[::,n] for n,key  in enumerate(["x","y","u","v"])}}

def get_grid(list_of_lines):
    return {"grid":{key:np.array([[i for i in i.split(" ") if len(i)>0] for i in [i.replace("\n","") for i in list_of_lines][1:] if len(i)>0],dtype="object").astype(float)[::,n] for n,key  in enumerate(["x","y","u","v"])}}

def get_RE(list_of_lines):
   re = {'RE':float(list_of_lines[0].split(" ")[0])}
   print(re)
   return re

def run_lensmodel(modeling_path,run_name):
    #run_name = "model_run"
    try:
        path_to_run = os.path.join(modeling_path,f"{run_name}.dat")
        where_we_are = os.getcwd()
        path_to_lensmodel = os.path.join(os.getcwd(),"time_scales/lensmodel/lensmodel")
        st = os.stat(path_to_lensmodel)
        # Get permissions and make the file executable
        os.chmod(path_to_lensmodel, st.st_mode | stat.S_IEXEC)
        # Change the current working directory to work_pad
        os.chdir(os.path.join(where_we_are,"time_scales/lensmodel"))
        with open(path_to_run, 'r') as f:
            proc = subprocess.Popen([path_to_lensmodel], stdin = f, stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
            proc.wait()
        os.chdir(where_we_are)
    except KeyboardInterrupt:
        os.chdir(where_we_are)
        #print("Execution stopped by the user.")







def full_modeling_result(modeling_path,model_setup):
    model_result = {}
    for step in ["step_1","step_2","final_step","kappa_gamma","critical_caustic","grid","RE"]: #only step+"kappa_gamma","critical_caustic","grid"
        file_name = f"{step}.dat"
        if "step" in step:
            func = get_result_lensmodel
            if f"{step}.best" in os.listdir(modeling_path):
                file_name = f"{step}.best"
        elif step=="kappa_gamma":
            func = get_kappa_gamma
        elif step=="critical_caustic":
            func = get_critical_caustic
        elif step=="grid":
            func = get_grid
        elif step=="RE":
            func =get_RE
        if file_name not in  os.listdir(modeling_path):
            continue
        file = open(os.path.join(modeling_path,file_name), "r")
        list_of_lines = file.readlines()
        model_result[step] = func(list_of_lines)
        [os.remove(os.path.join(modeling_path,i) )for i in os.listdir(modeling_path) if step in i]
    model_result["model_setup"] = {"model_setup":model_setup}
    return model_result

def save_dict_to_hdf5(dic, filename):
    with h5py.File(filename, 'w') as f:
        _save_dict_to_hdf5_recursive(f, '/', dic)

def _save_dict_to_hdf5_recursive(f, path, dic):
    for key, item in dic.items():
        if isinstance(item, dict):
            _save_dict_to_hdf5_recursive(f, path + key + '/', item)
        else:
            f[path + key] = item

def read_hdf5(filename):
    with h5py.File(filename, 'r') as f:
        data = _read_hdf5_recursive(f, '/')
    return data

def _read_hdf5_recursive(f, path):
    data = {}
    for key, item in f[path].items():
        if isinstance(item, h5py.Group):
            data[key] = _read_hdf5_recursive(f, path + key + '/')
        else:
            # Decode byte strings to regular strings
            if isinstance(item, h5py.Dataset) and isinstance(item[()], bytes):
                data[key] = item[()].decode('utf-8')  # Specify the appropriate encoding
            elif isinstance(item, h5py.Dataset) and isinstance(item[()], np.ndarray):
                data[key] = [x if type(x)==np.float64 else x.decode('utf-8')  for x in item[()] ] # Specify the appropriate encoding
            else:
                data[key] = item[()]
    return data
def compare_dicts(dict1, dict2):
    # Check if both dictionaries have the same keys
    if set(dict1.keys()) != set(dict2.keys()):
        return False
    
    # Check if the values for each key are the same
    false_key=[]
    for key in dict1.keys():
        if np.all(dict1[key] != dict2[key]):
            false_key.append(key)
    if len(false_key)>0:
        #print(false_key)
        return False
    return True


import numpy as np 
import pandas as pd
from copy import deepcopy
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes, mark_inset
from .utils import ModelNotFoundError
#q=1-d_r["p[4]"][i]
#theta_E=np.sqrt((1+q**2)/(2*q))*d_r["p[1]"][i]#theta_E_gravlens





class Result_Handler(
):  
    mass_models_parameters = {'SIS':3,"SIE":5,"POW":6,"SIS+shear":5,"SIE+shear":7,"POW+shear":8}
    def __init__(self,lensmodel_system: dict,reduced_chi=False,max_separation=0.01,look_for_best_model=False,older_version=False):
        
        if "mcmc_chain" in lensmodel_system.keys():
            #print(self.lensmodel_system["model_name"]) 
            self.lensmodel_system = {}
            self.lensmodel_system[lensmodel_system["model_name"]] = lensmodel_system
        else:
            self.lensmodel_system = lensmodel_system
        #agregar system_name to model setup? 
        self.older_version = older_version 
        self.max_separation = max_separation
        self.reduced_chi = reduced_chi
        self.models = list(self.lensmodel_system.keys())
        self.pandas_from_results = pd.concat([self.make_pandas_from_results(model).reset_index(drop=True) for model in self.models])
        self.pandas_model_stats = self.get_stats()
        self.acepted_models = [model for model, sep in self.pandas_model_stats[["model_name","max(delta_images)"]].values if ((np.max(sep) <= self.max_separation)& (np.min(sep) != 0)) ]
        self.can_end = (len(self.acepted_models)) > 0
        if self.can_end:
            self.current_best = self.pandas_model_stats[[i in self.acepted_models for i in self.pandas_model_stats.model_name]].sort_values('log_Chi').iloc[0]
            #self.pandas_model_stats[[i in self.acepted_models for i in self.pandas_results.T["n_model"]]].T.loc["log_Chi"].astype(float).idxmin()                      
        else:
            self.current_best = self.pandas_model_stats.sort_values('log_Chi').iloc[0]
        self.current_best_n_model = self.current_best.model_name                   
        #self.panda_stats,self.pandas_full = self.make_pandas_result()
        #self.pandas_full_model_result = 
    
    def get_model_from_pandas(self,model:str=None):
        """model str"""
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        return self.pandas_from_results[self.pandas_from_results.model_name==model]
    
    def get_necessary_to_mcmc(self,model:str=None):
        """model str"""
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        dic_model = self.get_model(model=model)
        stats=self.pandas_model_stats[self.pandas_model_stats.model_name==model]
        return dic_model,stats.iloc[0]
    
    def get_model(self,model:str=None):
        """model str"""
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        return self.lensmodel_system[model]
    
    def get_lens_params(self,model:str=None):
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        local_model = self.get_model(model)
        print("mass_distribution:",local_model.get("model_setup").get("model_setup").get("mass_distribution"))
        return local_model.get('final_step').get('LENS PARMS')
    
    def make_pandas_from_results(self,model:str=None):
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        model_dic = self.get_model(model)
        model_k = model_dic["kappa_gamma"]["kappa_gamma"]
        model_h = model_dic['final_step']
        model_s = model_dic["model_setup"]["model_setup"]
        model_er = model_dic["RE"]["RE"]
        combined_df = pd.merge(pd.DataFrame(model_k), pd.DataFrame(model_h["images"]),left_on='x', right_on='ra_imput')
        #A = pd.DataFrame(list(model_h['LENS PARMS'].values()) * len(combined_df))
        #A.reset_index(drop=True)
        B =pd.DataFrame(list(model_h['SOURCE PARMS'].values()) * len(combined_df))
        C = pd.DataFrame([list(model_h['CHISQ'].values())]* len(combined_df),columns=['chis2_'+i for i in model_h['CHISQ'].keys()])
        D = pd.DataFrame([list(model_s.values())]* len(combined_df),columns=[i for i in model_s.keys()])
        pandas_comb = pd.concat([combined_df,B.reset_index(drop=True),C.reset_index(drop=True),D.reset_index(drop=True)], axis=1) 
        pandas_comb["magnitudes"] = model_s["magnitudes"]
        pandas_comb['astrometry_error'] = pandas_comb['astrometry_error'].values[0]
        pandas_comb['center_mass_error'] = np.array([*pandas_comb['center_mass_error'].values])
        pandas_comb['flux_error_by_image'] = pandas_comb['flux_error'].values.T[1][0]
        pandas_comb['flux_error'] = max(pandas_comb['flux_error_by_image'])
        pandas_comb = pandas_comb.loc[:, ~pandas_comb.columns.duplicated()]
        pandas_comb["einstein_radii"] = [model_er]*len(pandas_comb["magnitudes"])
        pandas_comb["magnitudes_corrected"] = -2.5*np.log10(10**(-(np.abs(pandas_comb['flux_output'])+ np.min(pandas_comb["magnitudes"]))/2.5)/np.abs(pandas_comb["magnification"]))
        pandas_comb["delta_images"] = np.sqrt( (pandas_comb['ra_imput'] - pandas_comb["ra_output"])**2+ (pandas_comb['dec_imput'] - pandas_comb["dec_output"])**2) #sqrt delta images
        pandas_comb["model_name"] = [model]*len(pandas_comb["magnitudes"])
        pandas_comb['log_Chi'] = abs(np.log10(pandas_comb['chis2_tot']))
        pandas_comb["median_einstein_radii"] = [np.nan]*len(pandas_comb["magnitudes"])
        pandas_comb["std_einstein_radii"] = [np.nan]*len(pandas_comb["magnitudes"])
        pandas_comb["component"] = model_s["component"]
        pandas_comb["ra"] = model_s["ra"]
        pandas_comb["dec"] = model_s["dec"]
        if 'mcmc' in model_dic.keys():
            sampled_re = model_dic["mcmc"]['mcmc_chain']['RE'].values
            pandas_comb["median_einstein_radii"] = [np.median(sampled_re)]*len(pandas_comb["magnitudes"])
            pandas_comb["std_einstein_radii"] = [np.std(sampled_re)]*len(pandas_comb["magnitudes"])
        return pandas_comb[["component"] + [col for col in pandas_comb.columns if col not in ["component", "index"]]]

   #Chi = {key : self.get_chis2(key)['tot'] for key in self.models}
    #     log_Chi = {key : abs(np.log10(float(self.get_chis2(key)['tot']))) for key in self.models}
    #     pandas_results = pd.DataFrame([lensmodel_system,BIC,AICc,AIC,n_model,mass_models,mass_models_parameters,einstein_radii,rmse,max_delta_images,median_demag,std_demag,Chi,log_Chi]
    #                                   ,index=["name","BIC","AICc","AIC","n_model","mass_models","mass_models_parameters","einstein_radii","rmse","max(delta_images)","median_demag","std_demag","Chi","log_Chi"])
    #   
    
    def get_stats(self):
        pandas_stats = []
        relevant_keys = ["name","model_name",'mass_distribution','einstein_radii','median_einstein_radii','std_einstein_radii','flux_error','astrometry_error','center_mass_error', \
            'band_to_model','chis2_tot', 'chis2_pos','chis2_flux','log_Chi']
        if not self.older_version:
            relevant_keys += ['zl', 'zs', 'photometric_system', 'Telescope', 'instrument', 'zpt', 'lambda_cen']
        for model_name in self.models:
            panda_loc = self.pandas_from_results[self.pandas_from_results['model_name']==model_name]
            #print(model_name,max(panda_loc.delta_images))
            max_sep = max(panda_loc.delta_images)
            if max_sep > self.max_separation:
                max_sep = np.round(max(panda_loc.delta_images),2)
            median_demag =  np.median(panda_loc.magnitudes_corrected)
            std_demag =  np.std(panda_loc.magnitudes_corrected)
            #print(panda_loc[["model_name",'einstein_radius']].drop_duplicates().values)
            pandas_stats.append([*panda_loc[relevant_keys].drop_duplicates().values[0],max_sep,median_demag,std_demag])
        return pd.DataFrame(pandas_stats,columns=relevant_keys+["max(delta_images)",'median_demag','std_demag'])
    
    def make_plot(self, model=None, add_info=False, add_critical=False, save='',remove_axis=False, x_zoom_list=None, y_zoom_list=None, zoom_size=0.2, zoom_positions=None):
        """
        Creates a scatter plot with optional zoomed inset subplots.

        Parameters:
        - model: The model to plot.
        - add_info: Whether to add additional info text.
        - add_critical: Whether to add critical curves.
        - save: Filename to save the plot.
        - remove_axis: Whether to remove the main plot axes.
        - x_zoom_list: List of tuples defining x-limits for zoom regions, e.g., [(x1_min, x1_max), (x2_min, x2_max)].
        - y_zoom_list: List of tuples defining y-limits for zoom regions, e.g., [(y1_min, y1_max), (y2_min, y2_max)].
        - zoom_size: Size of the inset axes relative to the main plot (default is 0.2).
        - zoom_positions: List of tuples defining the position of each inset axes in [left, bottom, width, height] format.
                        If None, positions will be automatically arranged at the bottom.
        """
        model = model or self.current_best_n_model
        if model not in self.models:
            raise ModelNotFoundError(model, self.models)
        model_dic = self.lensmodel_system[model]
        ra_imput = model_dic['final_step']['images']['ra_imput']
        dec_imput = model_dic['final_step']['images']['dec_imput']

        ra_output = model_dic['final_step']['images']['ra_output']
        dec_output = model_dic['final_step']['images']['dec_output']
        ra_lens_model, dec_lens_model = [], []
        ra_lens_in, dec_lens_in = [], []
        
        for i, key in enumerate(model_dic['final_step']['LENS PARMS'].keys()):
            ra_lens_model.append(model_dic['final_step']['LENS PARMS'][key]['p[1]'])
            dec_lens_model.append(model_dic['final_step']['LENS PARMS'][key]['p[2]'])
            ra_lens_in.append(model_dic["model_setup"]["model_setup"]['lens_ra'][i])
            dec_lens_in.append(model_dic["model_setup"]["model_setup"]['lens_dec'][i])
        components = model_dic["model_setup"]['model_setup']['component']
        num_zooms = 0
        if x_zoom_list and y_zoom_list:
            if len(x_zoom_list) != len(y_zoom_list):
                raise ValueError("x_zoom_list and y_zoom_list must have the same length.")
            else:
                num_zooms = len(x_zoom_list)
        fig, ax = plt.subplots(1+num_zooms, 1, figsize=(20, 15*(1+num_zooms)))
        ax = np.atleast_1d(ax)
        ax[0].scatter(ra_imput, dec_imput, label="Input Coordinates", facecolors='none', edgecolors='blue')
        ax[0].scatter(ra_output, dec_output, label="Output Coordinates", color="g", alpha=0.7)
        ax[0].scatter(ra_lens_model, dec_lens_model, label="Output Coordinates Lens", color="k", alpha=0.7)
        ax[0].scatter(ra_lens_in, dec_lens_in, label="Input Coordinates Lens", facecolors='none', edgecolors='red')

        # Add text labels
        for i, v in enumerate(components):
            ax[0].text(ra_imput[i], dec_imput[i], v, fontsize=40, alpha=0.8, horizontalalignment="right")

        # Add critical curves if requested
        if add_critical:
            critical = model_dic["critical_caustic"]["critical_caustic"]
            x = critical["x"]
            u = critical["u"]
            v = critical["v"]
            y = critical["y"]
            step = np.argmax(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
            ax[0].plot(x[:step+1], y[:step+1], label="Critical Curves", alpha=0.5)
            ax[0].plot(u[:step+1], v[:step+1], label="Caustic Curves", alpha=0.5)

        # Add additional info text if requested
        if add_info:
            stats = self.pandas_model_stats[self.pandas_model_stats.model_name == model]
            info_text = rf"Mass distribution:{stats.mass_distribution.values[0]}"+"\n"+rf"$max(\Delta images) = {stats['max(delta_images)'].values[0]:.3f}$"+"\n"+fr"$\chi^2 (total) = {stats['chis2_tot'].values[0]}$"

            ax[0].text(
                ax[0].get_xlim()[0] - ax[0].get_xlim()[1] * 0.2,
                ax[0].get_ylim()[0] + ax[0].get_ylim()[1] * 0.11,
                info_text,
                fontsize=20
            )

        ax[0].set_ylabel(r"$\Delta \delta \quad [\mathrm{arcsec}]$")
        ax[0].set_xlabel(r"$\Delta \alpha \quad [\mathrm{arcsec}]$")
        ax[0].invert_xaxis()
        ax[0].legend(loc="upper left", bbox_to_anchor=(1.05, 0.75), borderaxespad=0)  # Legend outside
        
        # Remove axes if requested
        if remove_axis:
            for spine in ax.spines.values():
                spine.set_visible(False)

            # Remove ticks
            ax[0].set_xticks([])
            ax[0].set_yticks([])

            # Optionally remove tick labels
            ax[0].set_xticklabels([])
            ax[0].set_yticklabels([])
        if save:
            plt.savefig(f"{save}.jpg")
            plt.close()
        else:
            plt.show()
        
        # # Handle zoomed inset plots
        # if x_zoom_list and y_zoom_list:
        #     if len(x_zoom_list) != len(y_zoom_list):
        #         raise ValueError("x_zoom_list and y_zoom_list must have the same length.")

        #     num_zooms = len(x_zoom_list)

            
        #     # Define default positions if not provided
        #     if not zoom_positions:
        #         # Arrange zooms horizontally at the bottom
        #         zoom_positions = []
        #         spacing = 0.05  # Spacing between insets
        #         total_width = num_zooms * zoom_size + (num_zooms - 1) * spacing
        #         start_x = 0.5 - total_width / 2  # Center the zooms
        #         for i in range(num_zooms):
        #             zoom_positions.append([start_x + i * (zoom_size + spacing), 0.05, zoom_size, zoom_size])

        #     for i in range(num_zooms):
        #         zoom_ax = inset_axes(
        #             ax,
        #             width=f"{zoom_size * 100}%",  # width as a percentage of parent_bbox
        #             height=f"{zoom_size * 100}%",
        #             loc='lower center',
        #             bbox_to_anchor=(zoom_positions[i][0], zoom_positions[i][1], zoom_positions[i][2], zoom_positions[i][3]),
        #             bbox_transform=ax.transAxes,
        #             borderpad=1
        #         )

        #         # Scatter plots in zoomed inset
        #         zoom_ax.scatter(ra_imput, dec_imput, facecolors='none', edgecolors='blue')
        #         zoom_ax.scatter(ra_output, dec_output, color="g", alpha=0.7)
        #         zoom_ax.scatter(ra_lens_model, dec_lens_model, color="k", alpha=0.7)
        #         zoom_ax.scatter(ra_lens_in, dec_lens_in, facecolors='none', edgecolors='red')

        #         # Set limits for zoom
        #         zoom_ax.set_xlim(x_zoom_list[i])
        #         zoom_ax.set_ylim(y_zoom_list[i])

        #         # Optional: Remove ticks from inset
        #         zoom_ax.set_xticks([])
        #         zoom_ax.set_yticks([])

        #         # Optional: Add rectangle on the main plot to indicate zoom area
        #         rect = plt.Rectangle(
        #             (x_zoom_list[i][0], y_zoom_list[i][0]),
        #             x_zoom_list[i][1] - x_zoom_list[i][0],
        #             y_zoom_list[i][1] - y_zoom_list[i][0],
        #             linewidth=1, edgecolor='red', facecolor='none', linestyle='--'
        #         )
        #         ax.add_patch(rect)

        #         # Optionally, connect the inset to the zoom area
        #         # mark_inset(ax, zoom_ax, loc1=2, loc2=4, fc="none", ec="0.5")

        # Save or show the plot
        
    
    # def make_plot(self,model=None,add_info=False,add_critial=False,save='',remove_axis=False):
    #     model = model or self.current_best_n_model
    #     if model not in self.models:
    #         raise ModelNotFoundError(model, self.models)
    #     model_dic = self.lensmodel_system[model]
    #     ra_imput =model_dic['final_step']['images']['ra_imput']
    #     dec_imput =model_dic['final_step']['images']['dec_imput']

    #     ra_output =model_dic['final_step']['images']['ra_output']
    #     dec_output = model_dic['final_step']['images']['dec_output']
    #     ra_lens_model,dec_lens_model = [],[]
    #     ra_lens_in,dec_lens_in = [],[]
    #     for i,key in enumerate(model_dic['final_step']['LENS PARMS'].keys()):
    #         ra_lens_model.append(model_dic['final_step']['LENS PARMS'][key]['p[1]'])
    #         dec_lens_model.append(model_dic['final_step']['LENS PARMS'][key]['p[2]'])#, model_dic['final_step']['LENS PARMS']['alpha1']['p[2]']
    #         ra_lens_in.append( model_dic["model_setup"]["model_setup"]['lens_ra'][i]),dec_lens_in.append( model_dic["model_setup"]["model_setup"]['lens_dec'][i])
    #     #ra_lens_in,dec_lens_in = model_dic["model_setup"]["model_setup"]['lens_ra'][0],model_dic["model_setup"]["model_setup"]['lens_dec'][0]

    #     fig, axes = plt.subplots(1, 1, figsize=(20, 10))
    #     axes.scatter(ra_imput,dec_imput,label="input coordinates",facecolors='none', edgecolors='blue')
    #     axes.scatter(ra_output,dec_output,label="output coordinates",color="g",alpha=0.7)
    #     axes.scatter(ra_lens_model,dec_lens_model,label="output coordinates lens",color="k",alpha=0.7)
    #     axes.scatter(ra_lens_in,dec_lens_in,label="input coordinates lens",facecolors='none', edgecolors='red')
    #     #print(component)
    #     [axes.text(ra_imput[i],dec_imput[i],v,fontsize=40,alpha=0.8,horizontalalignment="right") for i,v in enumerate(model_dic["model_setup"]['model_setup']['component'])]
        
    #     if add_critial:
    #         x = model_dic["critical_caustic"]["critical_caustic"]["x"]
    #         u = model_dic["critical_caustic"]["critical_caustic"]["u"]
    #         v = model_dic["critical_caustic"]["critical_caustic"]["v"]
    #         y = model_dic["critical_caustic"]["critical_caustic"]["y"]
    #         step = np.argmax(np.sqrt(np.diff(x)**2+np.diff(y)**2))
    #         axes.plot(x[:step+1],y[:step+1],label="critical curves",alpha=0.5)
    #         axes.plot(u[:step+1],v[:step+1],label="caustic curves",alpha=0.5)
    #     if add_info:
    #         stats=self.pandas_model_stats[self.pandas_model_stats.model_name==model]
    #         axes.text(axes.get_xlim()[0]-axes.get_xlim()[1]*0.2,axes.get_ylim()[0]+axes.get_ylim()[1]*0.11,rf"Mass distribution:{stats.mass_distribution.values[0]}"+"\n"+rf"$max(\Delta images) = {stats['max(delta_images)'].values[0]:.3f}$"+"\n"+fr"$\chi^2 (total) = {stats['chis2_tot'].values[0]}$",fontsize=20)
    #     axes.set_ylabel(r"$\Delta \delta \quad [\mathrm{arcsec}]$")
    #     axes.set_xlabel(r"$\Delta \alpha \quad [\mathrm{arcsec}]$")
    #     axes.invert_xaxis()
    #     axes.legend(loc="upper left", bbox_to_anchor=(1.05, 0.75), borderaxespad=0)  # Legend outside
    #     plt.tight_layout(rect=[0, 0, 0.85, 1])  # Adjust layout to make room for the legend
    #     if remove_axis:
    #         for spine in axes.spines.values():
    #             spine.set_visible(False)

    #         # Remove ticks
    #         axes.set_xticks([])
    #         axes.set_yticks([])

    #         # Optionally remove tick labels
    #         axes.set_xticklabels([])
    #         axes.set_yticklabels([])
        
    #     if save:
    #         plt.savefig(f"{save}.jpg")
    #         plt.close()
    #     else:
    #         plt.show()
    
    
    
    
    #return model_dic 
    
    # def make_pandas_result(self):
    #     pandas_stats = []
    #     pandas_full = []
    #     for model_name in self.models:
    #         #delta_images = self.get_delta_images(model_name)
    #         #max_delta_images = np.max(abs(delta_images))
    #         pandas_stats.append([model_name,max_delta_images])
        
    #     return pd.DataFrame(pandas_stats,columns=['model_name',"max_delta_images"]),pandas_full
        #self.delta_images = {i : self.get_delta_images(i) for i in self.models}
        #self.pandas_results = self.make_pandas_model_stats(self.models)
        
        #self.rmse = {i : self.get_rmse(i) for i in self.models}
        #self.mass_models = {i : self.get_model_setup(i)['mass_distribution'] for i in self.models}
        #self.data_constrain = {i : self.get_constrains(i) for i in self.models}
        #self.einstein_radii = {i : self.get_einstein_radii(i) for i in self.models}
        #self.free_parameters = {i : self.data_constrain[i] - self.mass_models_parameters[self.mass_models[i]] for i in self.models}
        #self.demag = {i : self.get_demag(i) for i in self.models}
        #self.chis2 = pd.DataFrame({i : self.get_chis2(i,reduced_chi=self.reduced_chi) for i in self.models})
        #self.chis2["reduced_chi"] = [self.reduced_chi]*len(pd.DataFrame({i : self.get_chis2(i,reduced_chi=self.reduced_chi) for i in self.models}))
        #self.BIC = {i : self.get_BIC(i) for i in self.models}
        #self.AIC = {i : self.get_AIC(i) for i in self.models}
        #self.AICc = {i : self.get_AIC(i,aicc=True) for i in self.models}
        #self.einstein_radii = {i : self.get_einstein_radii(i) for i in self.models}
        #self.JJV = {i : self.get_JJV(i) for i in self.models}
        
        # self.acepted_models = [key for key, sep in self.delta_images.items() if ((np.max(sep) <= self.max_separation)& (np.min(sep) != 0)) ]
        # self.can_end = (len(self.acepted_models)) > 0
        
        # if self.can_end:
        #     self.current_best =self.pandas_results.T[[i in self.acepted_models for i in self.pandas_results.T["n_model"]]].T.loc["log_Chi"].astype(float).idxmin()                      
        # else:
        #     self.current_best = self.pandas_results.loc["Chi"].astype(float).idxmin()
        # self.current_best_info = self.pandas_results[self.current_best]
        # self.current_best_n_model = self.current_best_info.n_model
        #self.best_current_model = self.get_model(self.current_best_info.n_model)   #nnn not sure about this 
        #self.best_current_model["model"] = self.current_best_info.n_model
        #self.pandas_best = self.pandas_results.loc[self.current_best]
    #def get_info_to_mcmc(self,model):
     #   model = self.get_model(model)
        
    # def get_stats(self,model=None):
    #     model = model or self.current_best_n_model
    #     if model not in self.models:
    #         raise ModelNotFoundError(model, self.models)
    #     return self.pandas_results.T[self.pandas_results.T.n_model==model]
    
   
    
    # def get_mcmc_info(self,model):
    #     model = model or self.current_best_n_model
    #     if model not in self.models:
    #         raise ModelNotFoundError(model, self.models)
    #     dic_model = self.get_model(model=model)
    #     basic_info = self.pandas_results.T[self.pandas_results.T.n_model==model].iloc[0].to_dict()
    #     basic_info.update({keys:dic_model["model_setup"]["model_setup"][keys] for keys in ["zl","zs"]})
    #     #basic_info.update(keys:self.pandas_results.T[pandas_results.T.n_model==model])
    #     return dic_model,basic_info

    
            
    # def get_JJV(self,model=None):
    #     #mmmm
    #     model = model or self.current_best_n_model
    #     if model not in self.models:
    #         raise ModelNotFoundError(model, self.models)
    #     model_h_ = deepcopy(self.get_model(model)["kappa_gamma"]["kappa_gamma"])
    #     dicts_ = deepcopy(self.get_model_setup(model))
    #     main_shape = np.array(dicts_['astrometry_error']).shape
    #     for key,i in dicts_.items():
    #         if np.array(i).shape!=main_shape:
    #             if isinstance(i,str):
    #                 dicts_[key] = np.array([i]*main_shape[0])
    #             else:
    #                 dicts_[key] = np.array(i*main_shape[0])
    #     model_h_.update(dicts_)
    #     model_h_.pop('lens_dec', None)
    #     model_h_.pop('lens_ra', None)
    #     return pd.DataFrame(model_h_)
    
    
    # def get_kappa_gamma(self,model):
    #     model_h = self.get_model(model)["kappa_gamma"]["kappa_gamma"]
    #     return model_h
    # def get_constrains(self,model):
    #     # based on https://iopscience.iop.org/article/10.1088/0004-637X/773/1/35/pdf 
    #     n,l = len(self.get_model_setup(model)['component']),len(self.get_model_setup(model)['lens_ra'])*2
    #     return 3*n -1 + l
    # def get_BIC(self,model,chikey="pos"):
    #     #Fotios K. Anagnostopoulos + 2019
    #     #BIC = -2ln(L) + kln(Ntot)
    #     # -2ln(L) == chi2  (TESTING THE DARK ENERGY WITH GRAVITATIONAL LENSING STATISTIC, Shuo Cao), prior uniforme (?)
    #     #k: number of parameters of the model 
    #     #Ntot number of contrictions o data of the model 
    #     return self.get_chis2(model,reduced_chi=False)[chikey] + self.mass_models_parameters[self.mass_models[model]] * np.log(self.data_constrain[model])
    # def get_AIC(self,model,chikey="pos",aicc=False):
    #     #https://en.wikipedia.org/wiki/Akaike_information_criterion
    #     #k: number of parameters of the model
    #     # -2ln(L) == chi2  (TESTING THE DARK ENERGY WITH GRAVITATIONAL LENSING STATISTIC, Shuo Cao), prior uniforme (?)
    #     # denotes the number of parameters
    #     k = self.mass_models_parameters[self.mass_models[model]]
    #     n = self.data_constrain[model]
    #     #n denotes the sample 
    #     AICc = 0 if aicc==False else (2*(k)**2+2*k)/(n-k-1)
    #     return self.get_chis2(model,reduced_chi=False)[chikey] +  AICc
    # def get_chis2(self,model,reduced_chi=False):
    #     model_h = self.get_model(model)['final_step']['CHISQ']
    #     if reduced_chi:
    #         return {key: values/self.free_parameters[model] for key,values in model_h.items()}
    #     return model_h
    # @staticmethod
    # def rmse(v1, v2):
    #         return np.sqrt(np.mean(np.sum((v1 - v2)**2)))
    
    # def get_rmse(self,model):
    #     model_h = self.get_model(model)['final_step']
    #     imput_x,imput_y = np.array(model_h['images']['ra_imput']),np.array(model_h['images']['dec_imput'])
    #     model_x,model_y = np.array(model_h['images']['ra_output']),np.array(model_h['images']['dec_output'])
    #     rms_x = Result_Handler.rmse(imput_x, model_x)
    #     rms_y = Result_Handler.rmse(imput_y, model_y)
    #     rms_radial = Result_Handler.rmse(np.hypot(model_x, model_y), np.hypot(imput_x, imput_y))
    #     return rms_radial
    
    # def get_demag(self,model):
    #     model_h = self.get_model_setup(model)
    #     #first we recover the new magnitudes from the out put results and then we apply the magnification correction
    #     magnitudes = model_h['magnitudes']
    #     model_f_out_put = np.abs(np.array(self.get_model(model)['final_step']['images']['flux_output']))
    #     new_magnitudes = -2.5*np.log10(model_f_out_put)+ np.min(magnitudes)
    #     magnification =  np.array(self.get_kappa_gamma(model)['magnification'])
    #     mag_corrected = -2.5*np.log10((10**(-new_magnitudes/2.5))/np.abs(magnification))
    #     return mag_corrected
    
    # # def get_best_model(self,max_separation=0.001,current_model=None,print_best=False,get_all=False,best_panda=False,only_good_models=False,**kwargs):
    # #     #
    # #     #pandas_results = pd.DataFrame({})
    # #     #if len(acepted_models)>0:
    # #      #   can_end = True
    # #       #  pandas_results = self.make_pandas_model_stats(acepted_models)
    # #        # current_best = pandas_results.loc["AIC"].astype(float).idxmax()
    # #        # if print_best:
    # #         #    print(f"The best model for {self.name} that have a separation bellow {max_separation} is {current_best}, it is the best model taking in consideration the separation and AIC")
    # #     #if get_all or len(acepted_models)==0 :
    # #     #acepted_models = 
    # #     self.pandas_results = self.make_pandas_model_stats(self.models)
    # #     self.current_best = self.pandas_results.loc["AIC"].astype(float).idxmax()
    # #     #if len(self.models)>0:
    # #      #   self.can_end = True
            
            
    # #         #print(f"Any model have a separation bellow {max_separation}, but the better one is {current_best} taking in consideration just AIC")
    # #     #if best_panda:
    # #      #   pandas_results = pandas_results[current_best]
    # #       #  if only_good_models and not can_end:
    # #        #     pandas_results = pd.DataFrame({})
    # #     #self.pandas_results = pandas_results
    # #     #self.current_best = current_best
    # #     #print(pandas_results[current_best].mass_models)
    # #     #return current_best,can_end,pandas_results

    # def make_pandas_model_stats(self,models,extra_fields=False):
    #     "given a list of models make a pandas data frame"
    #     lensmodel_system = {key: self.name for key in models}
    #     BIC = {key: self.BIC[key] for key in models}
    #     AICc = {key: self.AICc[key] for key in models}
    #     AIC = {key: self.AIC[key] for key in models}
    #     n_model = {key: key  for key in models}
    #     #except:
    #      #   n_model = {key: key.split("(")[0] for key in models}
    #     mass_models = {key: self.mass_models[key] for key in models}
    #     mass_models_parameters = {key: int(self.mass_models_parameters[self.mass_models[key]]) for key in models}
    #     einstein_radii = {key: self.einstein_radii[key] for key in models}
    #     demag = {key: self.demag[key] for key in models}
    #     median_demag = {key: np.median(demag[key]) for key in models}
    #     std_demag = {key: np.std(demag[key]) for key in models}
    #     rmse = {key: self.rmse[key] for key in models}
    #     max_delta_images = {key: np.max(self.delta_images[key]) for key in models}
    #     Chi = {key : self.get_chis2(key)['tot'] for key in self.models}
    #     log_Chi = {key : abs(np.log10(float(self.get_chis2(key)['tot']))) for key in self.models}
    #     pandas_results = pd.DataFrame([lensmodel_system,BIC,AICc,AIC,n_model,mass_models,mass_models_parameters,einstein_radii,rmse,max_delta_images,median_demag,std_demag,Chi,log_Chi]
    #                                   ,index=["name","BIC","AICc","AIC","n_model","mass_models","mass_models_parameters","einstein_radii","rmse","max(delta_images)","median_demag","std_demag","Chi","log_Chi"])
    #     pandas_results = pandas_results.rename(columns={key:f"{key}({mass_model})" for key,mass_model in mass_models.items()})
    #     if extra_fields:
    #         return pandas_results
    #     return  pandas_results
        
    # #    def make_pandas_model_stats(self,models):
    # #     "given a list of models make a pandas data frame"
    # #     BIC = {key: self.BIC[key] for key in models}
    # #     AICc = {key: self.AICc[key] for key in models}
    # #     AIC = {key: self.AIC[key] for key in models}
    # #     mass_models = {key: self.mass_models[key] for key in models}
    # #     mass_models_parameters = {key: int(self.mass_models_parameters[self.mass_models[key]]) for key in models}
    # #     einstein_radii = {key: self.einstein_radii[key] for key in models}
    # #     delta_images = {key: self.delta_images[key] for key in models}
    # #     rmse = {key: self.rmse[key] for key in models} 
    # #     demag = {key: self.demag[key] for key in models} 
    # #     pandas_results = pd.DataFrame([mass_models,BIC,AICc,AIC,mass_models_parameters,einstein_radii,delta_images,rmse,demag],\
    # #                                   index=["mass_models","BIC","AICc","AIC","mass_models_parameters","einstein_radii","delta_images","rmse","demag"])
    # #     pandas_results = pandas_results.rename(columns={key:f"{key}" for key,mass_model in mass_models.items()})
    # #     return  pandas_results
    # def get_einstein_radii(self,model):
    #     return self.get_model(model)["RE"]["RE"]
    
    
    
    
    
    # def get_pandas(self,model):
    #     return 


import numpy as np
from scipy.ndimage import gaussian_filter
from .maps_utils import map2d_to_stats,renormalize_mean


dict_maps = {"alpha_0.1":  "/home/felipe/work/microlensing_timescale_qso_lensing/analysis/microlensing_maps/Mapas_magnification",
             "alpha_0.2":  "/home/felipe/work/microlensing_timescale_qso_lensing/analysis/microlensing_maps/Mapas_02"}


def maps_to_stats(name,component,rs_lday,pix,mappix,num_bins=100,bins_limit=3,rs_times = [0.3,1,2],dict_maps=dict_maps):
   """_summary_

   Args:
       name (_type_): _description_
       component (_type_): _description_
       rs_lday (_type_): _description_
       pix (_type_): _description_
       mappix (_type_): _description_
       num_bins (int, optional): _description_. Defaults to 100.
       bins_limit (int, optional): _description_. Defaults to 3.
       factors (list, optional): _description_. Defaults to [0.3,1,2].
       dict_maps (_type_, optional): _description_. Defaults to dict_maps.

   Returns:
       _type_: _description_
   """
   mag_maps = {}
   for key,path in dict_maps.items():
      map_1d_path = f'{path}/{name}{component}/magmap.dat'
      map_1d = np.loadtxt(map_1d_path)
      map_2d = np.reshape(map_1d, (int(mappix), int(mappix)))
      n_factor_results = {}
      for rs_time in rs_times:
         result = map2d_to_stats(map_2d,pix,rs_lday,num_bins=num_bins,bins_limit = bins_limit,rs_times = rs_time)
         n_factor_results[str(rs_time)] = result
      mag_maps[key] = n_factor_results
   return mag_maps
      
# def map_to_stat_full(name,component,rs_lday,pix,mappix):
#    print(f'Doing {name} {component}')
#    maps_alpha_01 = "/home/felipe/work/microlensing_timescale_qso_lensing/analysis/microlensing_maps/Mapas_magnification"
#    maps_alpha_02 = "/home/felipe/work/microlensing_timescale_qso_lensing/analysis/microlensing_maps/Mapas_02"
   
#    #magnification_map_path = f'microlensing_maps/{path_mapas}/{name}{component}/magmap.dat'
#    mag_maps = {}
#    for key,pats in zip(["alpha_0.1","alpha_0.2"],[maps_alpha_01,maps_alpha_02]):
#       map_1d_path = f'{pats}/{name}{component}/magmap.dat'
#       map_1d = np.loadtxt(map_1d_path)
#       map_2d = np.reshape(map_1d, (int(mappix), int(mappix)))
#       map_2d_norm = map_2d/np.mean(map_2d)
#       mag_map_2d_norm = -2.5*np.log10(map_2d_norm)
#       S = {}
#       for factor in [0.3,1,2]:
#          mag_map_2d_norm_conv = gaussian_filter(mag_map_2d_norm, [pix*rs_lday*factor,pix*rs_lday*factor],mode='mirror')
#          num_bins = 100
#          bins = np.linspace(-3, 3, num_bins + 1)
#          counts, bin_edges = np.histogram(mag_map_2d_norm_conv.ravel(), bins=bins)
#          pmf = counts / counts.sum()
#          bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
#          pdf = counts / (counts.sum() * np.diff(bin_edges))
#          expected_pmf = sum(pmf * bin_centers)
#          mean = np.sum(bin_edges * np.r_[pmf, pmf[-1]])
#          abs_mean = np.sum(abs(bin_edges) * np.r_[pmf, pmf[-1]])
         
#          #cumulative = np.cumsum(pmf)
#          # Find the median index (the first outcome where cumulative probability is at least 0.5)
#          #median = outcomes[np.where(cumulative >= 0.5)[0][0]]
         
#          var_pmf = sum(pmf * bin_centers**2) - expected_pmf**2
#          expected_pmf_abs = sum(pmf * abs(bin_centers))
#          var_pmf_abs = sum(pmf * abs(bin_centers)**2) - expected_pmf_abs**2
#          mean_pos_mag = renormalize_mean(pmf,bin_centers,bin_centers>0)
#          mean_neg_mag = renormalize_mean(pmf,bin_centers,bin_centers<0)
#          prob_mag_less_than_one = sum(pmf[bin_centers<-1])
#          prob_mag_less_than_m_032 = sum(pmf[bin_centers<-0.32])
#          prob_mag_more_than_05 = sum(pmf[bin_centers>0.5])
#          prob_mag_more_than_m_032 = sum(pmf[bin_centers>0.32])
         
#          #renormalize_mean(system_info[image]['pmf'],system_info[image]['bin_centers'],system_info[image]['bin_centers']>0)
#          small = {
#             "bin_centers": bin_centers, "pmf": pmf, "pdf": pdf,
#             "width": np.diff(bin_edges), "dx": np.diff(bin_edges)[0],
#             "expected_pmf": expected_pmf, "var_pmf": var_pmf,
#             "expected_pmf_abs": expected_pmf_abs, "var_pmf_abs": var_pmf_abs,
#             "bin_edges": bin_edges,"mag_map_2d_norm":mag_map_2d_norm,"mag_map_2d_norm_conv": mag_map_2d_norm_conv,
#             "factor":factor,'rs_lday':rs_lday,"mean_pos_mag":mean_pos_mag,"mean_neg_mag":mean_neg_mag,"prob_mag_less_than_one":prob_mag_less_than_one,"prob_mag_less_than_m_032":prob_mag_less_than_m_032,
#             "prob_mag_more_than_05":prob_mag_more_than_05,"prob_mag_more_than_m_032":prob_mag_more_than_m_032,"mean":mean}
#          S[factor] = small
#       mag_maps[key] = S
#    return mag_maps


import numpy as np
import scipy.odr as odr
from matplotlib.table import Table
from matplotlib.font_manager import FontProperties
import matplotlib.pyplot as plt

def linear_func(params, x):
    m, b = params
    return m * x + b

def evaluate_with_uncertainties(x_new, x_new_err, slope, intercept, slope_err, intercept_err):
    # Predicted y value
    y_new = slope * x_new + intercept
    
    # Uncertainty in predicted y value (including uncertainty in x_new)
    y_err_new = np.sqrt(
        (x_new * slope_err)**2 +              # Uncertainty due to slope
        intercept_err**2 +                    # Uncertainty due to intercept
        (slope * x_new_err)**2                # Uncertainty due to x_new
    )
    
    return y_new, y_err_new
def make_linear_fit(x,y,y_error=None,x_error=None,plot=False,names=[],title="title",save="",add_table=False,xlabel=r"$r_{s}$(thin disk) [lday]",ylabel=r"$r_{s}$(microlensing) [lday]"):
    linear_model = odr.Model(linear_func)
    data = odr.RealData(x, y, sx=x_error, sy=y_error)
    # Set initial guesses for slope (m) and intercept (b)
    beta0 = [1.0, 1.0]  # initial guess for [slope, intercept]
    # Create ODR object and run the fitting
    odr_fit = odr.ODR(data, linear_model, beta0=beta0)
    output = odr_fit.run()
    # Print the results
    slope, intercept = output.beta
    slope_err, intercept_err = output.sd_beta
    print(f"Slope: {slope} ± {slope_err}")
    print(f"Intercept: {intercept} ± {intercept_err}")
    # Generate the fit line
    mfc = "k"
    if add_table:
        mfc = 'white'
    if plot:
        x_fit = np.linspace(min(x)*0.95, max(x)*1.1, 100)
        y_fit = linear_func((slope, intercept),x_fit)

        # Propagate uncertainty across the entire fit line
        y_fit_upper = linear_func((slope + slope_err, intercept + intercept_err),x_fit)
        y_fit_lower = linear_func((slope - slope_err, intercept - intercept_err),x_fit)

        fig, ax = plt.subplots(figsize=(20, 10))
        ax.errorbar(x, y, yerr=y_error, xerr=x_error, fmt='o', markersize=10, mfc=mfc, mec='k', capsize=5,ecolor="k")
        #.errorbar(x, y,xerr=x_error,yerr=y_error, fmt='o',c="k", capsize=2, s=500, facecolors='none', edgecolors='blue')
        ax.plot(x_fit,y_fit, linewidth=1, color='r', label=r"Linear regression", ls='--', zorder=0)
        ax.fill_between(x_fit,y_fit_lower,y_fit_upper,
                                color="red",label=r'$\sigma$ uncertainty',alpha=0.3)
        
        if add_table:
            if len(names) ==0:
                names = [str(i+1) for i in range(len(x))]
            
            names = {i+1:name for i,name in enumerate(names)}
            for i, (num,name) in enumerate(names.items()):
                ax.text(x[i], y[i], str(num), fontsize=8, ha='center', va='center_baseline',zorder=10)
            # Create a table for the definitions
            table_data = [[str(num), fr"{definition}|$r_{{td}}=${x[num-1]:.4f} $\pm$ {x_error[num-1]:.4f} vs $r_{{m}}=${y[num-1]:.4f} $\pm$ {y_error[num-1]:.4f}"] for num, definition in zip(names.keys(), names.values())]

            # Add the table on the side
            
            table = Table(ax, bbox=[1.1, 0., 0.5, 1])  # Adjust bbox to control position and size of the table
            for i, (num, definition) in enumerate(table_data):
                table.add_cell(i, 0, width=0.2, height=0.1, text=num, loc='center')
                table.add_cell(i, 1, width=0.5, height=0.1, text=definition.split("|")[0], loc='left')
                table.add_cell(i, 2, width=1.5, height=0.1, text=definition.split("|")[1], loc='left')

            # Add the table to the plot
            ax.add_table(table)

            # Show plot 
            #ax.plot([12,17], [12,17], color='grey', label=r'$R_s=R_s$',alpha=0.8,ls='--',zorder=3)
            #plt.xlim(12.6,16)
            #plt.ylim(14,17.5)
            #plt.text(15,13,size=20)
        #
        ax.set_xlabel(xlabel, fontsize=20)
        ax.set_ylabel(ylabel, fontsize=20)
        ax.tick_params(axis='both', which='major', labelsize=20)
        ax.legend(fontsize=20)
        ax.set_xlim(x_fit[[0,-1]])
        ax.set_ylim(y_fit[[0,-1]])
        ax.set_title(title)
        plt.tight_layout()
        if len(save)>0:
            plt.savefig(f"{save}.png",bbox_inches='tight')
        plt.show()
    return slope, intercept, slope_err, intercept_err


from astropy.cosmology import FlatLambdaCDM
#,FlatwCDM,Flatw0waCDM,FlatJBP,FlatBA,FlatINTR,FlatCARD
from astropy import constants as const
from scipy import interpolate
from astropy import units as u
import sympy as sp
from sympy import symbols, diff
import statsmodels.api as sm
import numpy as np
import pandas as pd

def noleter(L):
    if type(L) != str:
        l=[]
        for ii in L:
            for i in [chr(i) for i in range(ord('A'),ord('Z')+1)]:
                    ii=ii.replace(i,"").replace("†","")
            l.append(ii)
        L=l
    else:
        for i in [chr(i) for i in range(ord('A'),ord('Z')+1)]:
            L=L.replace(i,"").replace("†","")
    return L
def codigoxd(txt):
    if txt=="None":
        return None
    xd=[]
    for i in range(len(txt)):
        if ","==txt[i] or "["==txt[i] or "]"==txt[i]:
            xd.append(i)
            #print(i)
    return np.array([float(txt[xd[ii]:xd[ii+1]].replace("[","").replace(",","")) for ii in range(0,len(xd)-1)])

class microlensingtimescale(object):
    cms = u.cm / u.s#cm/s
    G=const.G.to("cm3/s2solMass").value#cm/(s2SM)
    c=const.c.to("cm/s").value
    def __init__(self,cosmologia):
#         self.target = target   #may not need this
        self.cosmologia = cosmologia
    
    def dsig_pec_t(self,z):
        import statsmodels.api as sm
        import numpy as np
        x = np.array([0.,.5,1.0,2.0,3.0])
        y = np.array([179.544800,162.386536,144.020065,114.992302,95.906265])
        # include constant in ols models, which is not done by default
        x = sm.add_constant(x)
        model = sm.OLS(y,x)
        results = model.fit()
        y0,m=results.params
        dy0,dm=results.bse#
        y01,m1=results.params-results.bse
        y02,m2=results.params+results.bse
        yb=z*m+y0
        ymax,ymin=z*m1+y01,z*m2+y02
        yerror=abs((ymax-ymin)/2)
        return yb,yerror

    def dRe(self,zl,zs,M=0.3,dM=0.03,c=c,G=G): #radio einstein
        #zl lente
        #zs source
        #M solar mass es 0.3 segun paper profe error de esto?
        cos=self.cosmologia
        L={"Da_s":cos.angular_diameter_distance(zs).to(u.cm).value,
        "Da_l":cos.angular_diameter_distance(zl).to(u.cm).value,
        "Da_ls":cos.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value,"M":M,"G":G,"c":c}
        [Da_s,Da_l,Da_ls,M,G,c]=symbols([*L], real=True)
        v=[Da_s,Da_l,Da_ls,M,G,c]
        f=Da_s*(4*G*M*Da_ls/(Da_s*Da_l*c**2))**(0.5)
        F=f.evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))})
        df=abs(diff(f, M).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dM
        return float(F),float(df)#en cm
    
    def dRs(self,m,dm,z,zpt,lam,c=c,G=G,i=np.pi/3):
        #ya esta haciendo el cambio de variable 
        cos=self.cosmologia
        lamI=0.8140#micrometros
        L={"h":cos.H0.value/100,"r_h":c/cos.H0.to("cm/Mpc s").value,
        "Da":cos.angular_diameter_distance(z).value,"i":np.pi/3,"lamda":lam,"zpt":zpt,"m":m}
        [h,r_h,Da,i,lamda,zpt,m]=symbols([*L], real=True)
        v=[h,r_h,Da,i,lamda,zpt,m]
        f=((3.4*10**(15))/sp.sqrt(sp.cos(i)))*(Da/r_h)*((lamda)**(3/2))*((zpt/3631)**(0.5))*(10**(-0.2*(m-19)))*(h**(-1))
        df=abs(diff(f, m).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dm*(lamI/ L["lamda"])**(4/3)
        #https://iopscience.iop.org/article/10.3847/1538-4357/aac9bb/pdf equation (3)
        F=f.evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))})*(lamI/ L["lamda"])**(4/3)
        return float(F),float(df)
    
    def dsig_d_s(self,teta_e,dteta_e,zl,zs,c=c):#velocity dispersion of the stars in the lens galaxy
        #teta_e radio de einstein en arcsecons
        #zl,zs redshif fuente y lente
        #teta_e en arcsec(")
        cos=self.cosmologia
        L={"teta_e":teta_e*u.arcsec.to("rad"),"c":(c*u.cm/u.s).to("km/s").value,
        "Da_s":cos.angular_diameter_distance(zs).value,
        "Da_l":cos.angular_diameter_distance(zl).value,
        "Da_ls":cos.angular_diameter_distance_z1z2(zl,zs).value,"pi":np.pi}
        [teta_e,c,Da_s,Da_l,Da_ls,pi]=symbols([*L], real=True)
        v=[teta_e,c,Da_s,Da_l,Da_ls,pi]
        f=0.5*c*sp.sqrt(teta_e*Da_s/(pi*Da_ls))
        dteta_e=dteta_e*u.arcsec.to("rad")
        df=abs(diff(f, teta_e).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dteta_e
        F=f.evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))})
        return  float(F),float(df)
    
    def dVe(self,teta_e,dteta_e,zl,zs,v_CMB=369,dvcmb=0.9): # separation redshift lens redshift source peculiar velocity lens peculiar velocity source
        #teta_e radio de einsteint en arcsec(") y dteta_e error
        #v_cmb km/s#buscar esto en el paper de hinshaw 2009 #f 369.0 ± 0.9 km s−1   
        cos=self.cosmologia
        p_l,dp_l=self.dsig_pec_t(zl)
        p_s,dp_s=self.dsig_pec_t(zs)
        v_d,dv_d=self.dsig_d_s(teta_e,dteta_e,zl,zs)
        L={"Da_s":cos.angular_diameter_distance(zs).to(u.km).value,
        "Da_l":cos.angular_diameter_distance(zl).to(u.km).value,
        "Da_ls":cos.angular_diameter_distance_z1z2(zl,zs).to(u.km).value,
        "p_l":p_l,"p_s":p_l,"v_d":v_d,"v_CMB":v_CMB,"zl":zl,"zs":zs}
        #p_l velocidad peculiar lente
        #p_s velocidad peculiar fuente
        #v_d velocidad de dispercion
        [Da_s,Da_l,Da_ls,p_l,p_s,v_d,v_CMB,zl,zs]=symbols([*L], real=True)
        v=[Da_s,Da_l,Da_ls,p_l,p_s,v_d,v_CMB,zl,zs]
        f=sp.sqrt(((p_l*Da_s/((1+zl)*(Da_l)))**2)+((p_s/(1+zs))**2)+((v_CMB*Da_ls/((1+zl)*Da_l))**2)+(2*(v_d*Da_s/((1+zl)*Da_l))**2))
        df=abs(diff(f, p_l).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dp_l
        +abs(diff(f, p_s).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dp_s
        +abs(diff(f, v_CMB).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dvcmb
        +abs(diff(f, v_d).evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))}))*dv_d
        F=f.evalf(subs={v[i]:L[str(v[i])] for i in range(len(v))})
        return float(F),float(df) 
def teta(nombre_sistema,pandas,columna="noleter"):
    #b valor de p[0]
    #e valor de p[4]
    #db error de p[0]
    #de error de p[4]
    #pad=work_pad+models+"/"+str(nombre_sistema)
    #p=pd.read_csv(pad+"/result/"+'finalparameters_'+str(nombre_sistema)+".csv")
    try:
        p=pandas[pandas[columna]==nombre_sistema]
    except:
        print("revisar columnas")
        return
    #print(p["p[1]"+"_f"].values[0],((p["up_p[1]"+"_f"])+abs(p["down_p[1]"+"_f"]))/2)
    try:
        b=p["p[1]"+"_f"].values[0]
        db=((p["up_p[1]"+"_f"])+abs(p["down_p[1]"+"_f"]))/2
        db=db.values[0]
    except:
        b=p["p[1]"+"_f"]
        db=((p["up_p[1]"+"_f"])+abs(p["down_p[1]"+"_f"]))/2
        db=db[0]
    try:
        e=p["p[4]"+"_f"].values[0]
        de=((p["up_p[4]"+"_f"])+abs(p["down_p[4]"+"_f"]))/2
        de=de.values[0]
    except:
        e=0
        de=0
    theta_E=np.sqrt((1+(1-e)**2)/(2*(1-e)))*b#theta_E_gravlens
    etheta_E=np.sqrt((1+(1-e)**2)/(2*(1-e)))*db+b*((((1-e)**2+1)/((1-e)**2))-2)/(2*np.sqrt(2)*np.sqrt((((1-e)**2+1)/((1-e)))))*de
    return float(theta_E),float(etheta_E),float(e)
def zpt_lam(nombre_sistema,pandas,columna="noleter"): 
    #['g', 'r',"V","R","I","H",'i','z','Y','F160W','F814W','F475X',"G","5GHz","B","Kp"]
    try:
        p=pandas[pandas[columna]==nombre_sistema]
    except:
        print("revisar columnas")
        return
    System,band=p[["Psystem","banda"]].values[0]
    #i,f814,I  (λ = 0.814 µm and zpt= 2409) mosquera
    if System=="AB":
        zpt=3631
    if System=="vega":
        if band in ["Kp"]:
            zpt=676.31
        if band in ["i","I","F814W"]:
            zpt=2409
        if band in ["R","F160W","5GHz","r"]:
            zpt=1132
        if band=="G":
            zpt=2847.56
        if band=="R":
            zpt=3989.43
    if band in ["i","I","F814W"]:
        lam=0.814#micrometros
    if band in ["Kp"]:
        lam=21125.25/10000
    if band in ["G"]:
        lam=0.672
    if band in ["R"]:
        lam=4750.64/10000
    if band in ["R","F160W","5GHz","r"]:
        lam=15435/10000
    if band in ["B"]:
        lam=4336.33/10000
    if band in ['F475X']:
        lam=4937.39/10000
    return zpt,lam
def Result_pandas(cosmologia,nombres=[],GPanda=[],RPanda=[]):
    #Rλ ∝ λ4/3
    Cms=microlensingtimescale(cosmologia)#calculos basados en una cosmologia especifica
    D=pd.DataFrame()
    D["name"]=nombres
    D['model']=np.array([RPanda[RPanda["name"]==i]['model'].values[0] for i in nombres])
    D[["zl","zs"]]=np.array([GPanda[GPanda["name"]==i][["z_l","z_s"]].values[0] for i in nombres])
    D["zl"][D["zl"]=="-"]=0.5*np.ones(len(D["zl"]=="-"))
    D["zs"][D["zs"]=="-"]=2*np.ones(len(D["zs"]=="-"))
    D["zl"]=[i.replace("?","") if type(i)==str else i for i in D["zl"].values]
    D["zs"]=[i.replace("?","") if type(i)==str else i for i in D["zs"].values]
    D[["Rex16cm","dRex16cm"]]=np.around(np.array([Cms.dRe(zl,zs) for zl,zs in D[["zl","zs"]].astype(float).values])*10**(-16),4)
    D[["m","dm"]]=[[np.mean(RPanda[RPanda["name"]==i]["corrected"]),1.2*np.std(RPanda[RPanda["name"]==i]["corrected"])] for i in nombres]
    #suponiendo 20% de error
    #D[["m","dm"]]=[[np.mean(RPanda[RPanda["name"]==i]["corrected"]),0.1*np.mean(RPanda[RPanda["name"]==i]["corrected"])] for i in nombres]
    D[["zpt","lam"]]=[zpt_lam(i,RPanda) for i in nombres]
    D[["Rsx15cm","dRsx15cm"]]=np.round(np.array([Cms.dRs(m,dm,z,zpt,lam) for m,dm,z,zpt,lam in D[["m","dm","zs","zpt","lam"]].astype(float).values])*10**(-15),4)
    D[["teta_e","dteta_e","e"]]=[teta(nombre_sistema,RPanda) for nombre_sistema in nombres]
    D[["Vexkm/s","dVexkm/s"]]=np.around(np.array([Cms.dVe(teta_e,dteta_e,zl,zs) for teta_e,dteta_e,zl,zs in D[["teta_e","dteta_e","zl","zs"]].astype(float).values]),4)
    D["te[yr]"]=np.around((D["Rex16cm"]*10**(16)/(D["Vexkm/s"]*u.km.to("cm")))*u.s.to("yr"),4)
    D["dte[yr]"]=np.around((abs((D["Rex16cm"]*10**(16)/((D["Vexkm/s"]*u.km.to("cm"))**2)))*D["dVexkm/s"]*u.km.to("cm")
    +abs(1/(D["Vexkm/s"]*u.km.to("cm")))*D["dRex16cm"]*10**(16))*u.s.to("yr"),4)
    D["ts[yr]"]=np.around((D["Rsx15cm"]*10**(15)/(D["Vexkm/s"]*u.km.to("cm")))*u.s.to("yr"),4)
    D["dts[yr]"]=np.around((abs((D["Rsx15cm"]*10**(15)/((D["Vexkm/s"]*u.km.to("cm"))**2)))*D["dVexkm/s"]*u.km.to("cm")
    +abs(1/(D["Vexkm/s"]*u.km.to("cm")))*D["dRsx15cm"]*10**(15))*u.s.to("yr"),4)
    return D
#cosM=FlatLambdaCDM(H0=72,Om0=0.3, Ob0=0.,Neff=3.04) #Mosquera
#Microlensing_Thomas=Result_pandas(cosM,nombres=resultadosThomas.name.drop_duplicates(),GPanda=datosThomas,RPanda=resultadosThomas)
def Result_pandasv2(cosmologia,nombres=[],GPanda=[],RPanda=[]):
    #Rλ ∝ λ4/3
    Cms=microlensingtimescale(cosmologia)#calculos basados en una cosmologia especifica
    D=pd.DataFrame()
    D["name"]=nombres
    nombres_sinletra=noleter(nombres)
    RPanda["noleter"]=noleter(RPanda.name)
    GPanda["noleter"]=noleter(GPanda.name)
    D['model']=np.array([RPanda[RPanda["noleter"]==i]['model'].values[0] for i in nombres_sinletra])
    D["zl"] = [float(str(GPanda[GPanda["noleter"]==name]["z_l"].drop_duplicates().values[0]).replace("-","0.5").replace("?","")) for name in nombres_sinletra]
    D["zs"] = [float(str(GPanda[GPanda["noleter"]==name]["z_s"].drop_duplicates().values[0]).replace("-","2.0").replace("?","")) for name in nombres_sinletra]
    D[["Rex16cm","dRex16cm"]]=np.around(np.array([Cms.dRe(zl,zs) for zl,zs in D[["zl","zs"]].astype(float).values])*10**(-16),4)
    D[["m","dm"]]=[[np.mean(RPanda[RPanda["noleter"]==i]["corrected"]),1.2*np.std(RPanda[RPanda["noleter"]==i]["corrected"])] for i in nombres_sinletra]
    #suponiendo 20% de error
    #D[["m","dm"]]=[[np.mean(RPanda[RPanda["name"]==i]["corrected"]),0.1*np.mean(RPanda[RPanda["name"]==i]["corrected"])] for i in nombres]
    D[["zpt","lam"]]=[zpt_lam(i,RPanda) for i in nombres_sinletra]
    D[["Rsx15cm","dRsx15cm"]]=np.round(np.array([Cms.dRs(m,dm,z,zpt,lam) for m,dm,z,zpt,lam in D[["m","dm","zs","zpt","lam"]].astype(float).values])*10**(-15),4)
    D[["teta_e","dteta_e","e"]]=[teta(nombre_sistema,RPanda) for nombre_sistema in nombres_sinletra]
    D[["Vexkm/s","dVexkm/s"]]=np.around(np.array([Cms.dVe(teta_e,dteta_e,zl,zs) for teta_e,dteta_e,zl,zs in D[["teta_e","dteta_e","zl","zs"]].astype(float).values]),4)
    D["te[yr]"]=np.around((D["Rex16cm"]*10**(16)/(D["Vexkm/s"]*u.km.to("cm")))*u.s.to("yr"),4)
    D["dte[yr]"]=np.around((abs((D["Rex16cm"]*10**(16)/((D["Vexkm/s"]*u.km.to("cm"))**2)))*D["dVexkm/s"]*u.km.to("cm")
    +abs(1/(D["Vexkm/s"]*u.km.to("cm")))*D["dRex16cm"]*10**(16))*u.s.to("yr"),4)
    D["ts[yr]"]=np.around((D["Rsx15cm"]*10**(15)/(D["Vexkm/s"]*u.km.to("cm")))*u.s.to("yr"),4)
    D["dts[yr]"]=np.around((abs((D["Rsx15cm"]*10**(15)/((D["Vexkm/s"]*u.km.to("cm"))**2)))*D["dVexkm/s"]*u.km.to("cm")
    +abs(1/(D["Vexkm/s"]*u.km.to("cm")))*D["dRsx15cm"]*10**(15))*u.s.to("yr"),4)
    return D
def mag_to_flux(mag):
    return 10 ** (-0.4 * mag)

def flux_to_mag(flux):
    return -2.5 * np.log10(flux)

def mean_magnitude_with_error(magnitudes, errors):
    # Convert magnitudes to fluxes
    fluxes = [mag_to_flux(mag) for mag in magnitudes]
    # Convert errors in magnitudes to errors in fluxes
    flux_errors = [(flux) * np.log(10) * 0.4 * (err+1e-5) for flux, err in zip(fluxes, errors)]

    # Calculate the weighted mean flux and its error
    weights = [1 / (err ** 2) for err in flux_errors]
    weighted_mean_flux = np.sum([flux * 1 for flux, weight in zip(fluxes, weights)]) / np.sum(1)
    weighted_mean_flux_error = np.sqrt(1 / np.sum(weights))

    # Convert the weighted mean flux and its error back to magnitude and error in magnitude
    mean_mag = flux_to_mag(weighted_mean_flux)
    mean_mag_error = weighted_mean_flux_error / (weighted_mean_flux * np.log(10) * 0.4)

    return mean_mag, mean_mag_error


import numpy as np 


def ra_dec_to_unit_vector(ra, dec):
    """
    Convert RA and Dec to a unit vector.
    
    Parameters:
    ra  -- Right Ascension in degrees
    dec -- Declination in degrees
    
    Returns:
    A 3-element numpy array representing the unit vector.
    """
    ra_rad = np.radians(ra)
    dec_rad = np.radians(dec)
    
    x = np.cos(dec_rad) * np.cos(ra_rad)
    y = np.cos(dec_rad) * np.sin(ra_rad)
    z = np.sin(dec_rad)
    
    return np.array([x, y, z])

def calculate_projections(ra0, dec0,cmb_ra=167.9,cmb_dec = -6.9):
    """
    Calculate the projections of the CMB dipole on East and North directions.
    
    Parameters:
    cmb_ra  -- RA of the CMB dipole in degrees (J2000)
    cmb_dec -- Dec of the CMB dipole in degrees (J2000)
    ra0     -- RA of the point of interest in degrees (J2000)
    dec0    -- Dec of the point of interest in degrees (J2000)
    
    Returns:
    A tuple with (v_north, v_east), the projections of the dipole on the North and East directions.
    """
    # Convert CMB dipole direction to unit vector
    v_cmb = np.tile(ra_dec_to_unit_vector(cmb_ra, cmb_dec)[:,np.newaxis], ra0.shape).T
    #print(v_cmb.shape)
    # Calculate the unit vector for North direction
    dec0_rad = np.radians(dec0)
    ra0_rad = np.radians(ra0)
    
    v_north = np.array([-np.sin(dec0_rad) * np.cos(ra0_rad),
                        -np.sin(dec0_rad) * np.sin(ra0_rad),
                        np.cos(dec0_rad)])
    #print(v_north.shape)
    # Calculate the unit vector for East direction
    v_east = np.array([-np.sin(ra0_rad),
                        np.cos(ra0_rad),
                        np.zeros(ra0_rad.shape)])
    #print(v_east.shape)
    # Project the CMB dipole onto the North and East directions
    projection_north = np.dot(v_cmb, v_north)
    #print(projection_north.shape)
    projection_east = np.dot(v_cmb, v_east)
    #print(projection_east.shape)
    return np.diag(projection_north), np.diag(projection_east)


"""Run `python -m qumas`.

Allow running qumas, also by invoking
the python module:

`python -m qumas`

This is an alternative to directly invoking the cli that uses python as the
"entrypoint".
"""

from __future__ import absolute_import

from qumas.cli import main

if __name__ == "__main__":  # pragma: no cover
    main(prog_name="qumas")  # pylint: disable=unexpected-keyword-arg



from uncertainties.core import AffineScalarFunc
from astropy.constants import c
from astropy.units import cm, day
from uncertainties import unumpy,ufloat
import numpy as np

def montecarlo_errors(x_data,x_error,N=10000):
    medians = []
    for _ in range(N):
    # Resample each data point from its normal distribution
        x_sampled = np.random.normal(x_data, x_error)
    # Compute the median of the resampled data
        medians.append(np.median(x_sampled))

    # Calculate the overall median and its uncertainty
    median_estimate = np.median(medians)
    median_uncertainty = np.std(medians)

    print(fr"Median = {median_estimate:.3f} $\pm$ {median_uncertainty:.3f}")
    return median_estimate,median_uncertainty


def lday_to_cm(lday_value):
    # Speed of light in cm/s
    speed_of_light_cm_per_s = 2.998e10
    # Number of seconds in a day
    seconds_per_day = 86400
    
    # Convert light-days to cm
    cm_value = lday_value * speed_of_light_cm_per_s * seconds_per_day
    return cm_value
def cm_to_lday(cm_value):
    # Convert cm to meters
    meters = cm_value * cm.to('m')
    # Convert meters to light days
    light_days = meters / (c.to('m/day').value)
    return light_days

def RxtoR2500(Rx,x,Rx_in_log=True,ref=2500):
    """
    Rx radio at x lambda
    x lambda in A
    ref is the ref labda use to to be will be the radius
    
    """
    if Rx_in_log:
        Rx=10**(Rx)
    return np.log10(Rx*((ref/x)**(4/3)))
def log10_to_number(rx):
    """_summary_

    Args:
        rx (_type_): rx is a tuple (rx,sigma_rx) log10

    Returns:
        _type_: numpy array(2,) with rx and sigma_rx non log10
    """
    return to_return(10**(unumpy.uarray(*rx)))

def r_x_to_r_y(rx,x,y=2500,log10=False):
    """
    rx: tuple (rx,sigma_rx) at x lambda with error
    x: lambda in A at which rx is measured
    y: lambda in A at which ry will be re-scaled
    return and (2,len(rx)) array with ry and error
    
    """
    if log10:
        return to_return(unumpy.log10(unumpy.uarray(*rx)*(y/x)**(4/3)))
    else:
        return to_return(unumpy.uarray(*rx)*(y/x)**(4/3))

def to_return(f):
    if isinstance(f,AffineScalarFunc):
        f= np.array([f.nominal_value,f.s])
    elif len(f.shape)>1:
        f= np.array([[i.nominal_value,i.s] for i in f[0].T])
    elif len(f.shape)==1:
        f= np.array([[i.nominal_value,i.s] for i in f])
    else:
        f = f.reshape(1,)[0]
        f= np.array([f.nominal_value,f.s])
    return f


"""Set up Application Logs

This module defines how the emitted application logs are handled and where
they are written/streamed.
The application logs are written in full details (ie with timestamps) to a file
and also streamed to the console in a more concise format.

# Console/Terminal Log:
    - We Stream Logs of INFO (and above) Level on Console's stderr
    - The rendered Log format is: <logger name>: <log level> <log message>

# Disk File Log:
    - we Write Logs of ALL Levels on a Disk File
    - The rendered Log format is: <timestamp> <logger name>: <log level> <log message>
    - The FILE_TARGET_LOGS, variable (see below), defines the path to the log file

Log Levels:
- CRITICAL
- ERROR
- WARNING
- INFO
- DEBUG

Usage:
    Do a 'from . import _logging' in the root __init__.py of your package and
    all submodules 'inherit' the logging configuration
"""

import logging

# for 'biskotaki' app/code
FILE_TARGET_LOGS = 'biskotaki.log'

#### FILE LOGGING
# set up logging to file for DEBUG Level and above
logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
    datefmt='%m-%d %H:%M',
    filename=FILE_TARGET_LOGS,
    filemode='w',
)

#### CONSOLE LOGGING
console = logging.StreamHandler()

### Handler which writes DEBUG messages or higher to the sys.stderr ###
console.setLevel(logging.DEBUG)
# console.setLevel(logging.INFO)

# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('').addHandler(console)


# Now, we can log to the root logger, or any other logger. First the root...
# logging.info('Blah blah')

# Now, define a couple of other loggers which might represent areas in your
# application:

# logger1 = logging.getLogger('myapp.area1')
# logger2 = logging.getLogger('myapp.area2')
# logger3 = logging.getLogger(__name__)

# logger1.debug('balh blah')
# logger1.info('balh blah')
# logger2.warning('balh blah')
# logger3.error('balh blah')



import glob
import numpy as np
import pandas as pd
from .LensmodelWrapper.utils import pandas_to_model
import os 

module_dir = os.path.dirname(os.path.abspath(__file__))
#list with the info for the different filters, this can be add to LensedQsoCensus

#filter_info =  pd.read_csv(f"{module_dir}/tables/filter_unique.csv")

class LensedQsoCensus:
    def __init__(self, path="Tables/photometry"):
        self.path_lens_tables = glob.glob(f"{module_dir}/{path}/*")
        #print(self.path_lens_tables)
        self.lens_census = self._do_full_lens_census()
        self.lens_census["year"] = [i[0:4] for i in self.lens_census["Bibcode"]]
        self.points = self.lens_census[["ra", "dec"]].values
        self.distances = np.linalg.norm(self.points[:, np.newaxis] - self.points, axis=2)

    def _do_full_lens_census(self):
        # Concatenate all tables into a single DataFrame
        return pd.concat(
            [pd.read_csv(file).assign(file=file) for file in self.path_lens_tables]
        ).reset_index(drop=True)

    def look_for_a_system(self, word, column="name", separation_limit=0.1):
        # Create a mask for non-NaN values and those containing the word
        #TODO add a filter for the columns i want to see 
        mask = self.lens_census[column].notna() & self.lens_census[column].astype("str").str.contains(str(word), regex=False)
        pandas_obj = self.lens_census[mask]
        
        if pandas_obj.empty:
            raise ValueError("Object not found in census.")
        
        # Calculate distance-based mask if needed
        if column in ["name", "z_s"]:
            idx = pandas_obj.index[0]
            close_mask = self.distances[idx] < separation_limit
            mask |= close_mask
        
        # Filter the data based on the combined mask
        filtered_data = self.lens_census[mask].dropna(axis=1, how='all')
        
        return filtered_data

    def unique_systems_count(self,get_unique_names=False,get_unique_years=False):
        # Initialize a mask for visited systems
        visited = np.zeros(len(self.lens_census), dtype=bool)
        count = 0
        names = []
        years = []
        # Vectorized approach to count unique systems
        for idx in range(len(self.lens_census)):
            if not visited[idx]:
                # Mark all systems within the separation limit as visited
                close_mask = self.distances[idx] < 0.1
                visited |= close_mask
                count += 1
                if get_unique_names:
                    names.append(self.lens_census.name[close_mask].drop_duplicates().values[0])
                if get_unique_years:
                    years.append([self.lens_census.name[close_mask].drop_duplicates().values[0],min(self.lens_census.year[close_mask].drop_duplicates().values.astype(int))])
        if get_unique_names:
            return names
        if get_unique_years:
            return years
        return count
    
    def hierarchical_selection(self,name):
        system = self.look_for_a_system(name).sort_values("year", ascending=False)
        pandas_s = np.array([[i,len(system[system["Bibcode"]==i])] for i in system.Bibcode.drop_duplicates().values])
        n_ = np.argmax(pandas_s[:,1].astype(int))
        data_to_model = system[system["Bibcode"]==system.Bibcode.drop_duplicates().values[n_]].dropna(axis=1, how='all').copy()
        if 'z_l' in system.columns and 'Bibcode' in system.columns:
            zl = [[zl, bibcode] for zl, bibcode in system[["z_l", "Bibcode"]].drop_duplicates().values if not pd.isnull(zl)][0]
            data_to_model[["z_l","Bibcode_zl"]] = [zl]* len(data_to_model)
        # Process the source redshift 'z_s'
        if 'z_s' in system.columns and 'Bibcode' in system.columns:
            zs = [[zs, bibcode] for zs, bibcode in system[["z_s", "Bibcode"]].drop_duplicates().values if not pd.isnull(zs)][0]
            data_to_model[["z_s","Bibcode_zs"]] = [zs] * len(data_to_model)
        data_to_model["known_names"] = [system.name.drop_duplicates().values]* len(data_to_model)
        data_to_model["can_be_modeled"] = [any([(("band" in col) and ("ima" in data_to_model["IS"].values)) for col in data_to_model.columns])] * len(data_to_model)
        return data_to_model
    # def pandas_for_model(self):
    #     pandas_to_be_use_in_model = pd.concat([pandas_to_model(self.hierarchical_selection(system_name)) for system_name in self.unique_systems_count(get_unique_names=True)]
    #     ).reset_index(drop=True)
    #     pandas_to_be_use_in_model['total_lens'] = pandas_to_be_use_in_model['total_lens'].fillna(0)
    #     return pandas_to_be_use_in_model

